import unittest

import numpy as np
import numpy.testing as npt
import wisdem.towerse.tower_props as tow
from wisdem.commonse.utilities import nodal2sectional

npts = 100
myones = np.ones((npts,))


class TestDiscretization(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params["section_height"] = np.arange(1, 5)
        self.params["diameter"] = 5.0 * np.ones(5)
        self.params["wall_thickness"] = 0.05 * np.ones(4)
        self.params["foundation_height"] = 0.0

    def testRefine2(self):
        mydis = tow.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns["z_param"], np.array([0.0, 1.0, 3.0, 6.0, 10.0]))
        npt.assert_array_equal(self.unknowns["z_full"], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]))
        npt.assert_array_equal(self.unknowns["d_full"], 5.0)
        npt.assert_array_equal(self.unknowns["t_full"], 0.05)

    def testFoundation(self):
        self.params["foundation_height"] = -30.0
        mydis = tow.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns["z_param"], np.array([0.0, 1.0, 3.0, 6.0, 10.0]) - 30.0)
        npt.assert_array_equal(self.unknowns["z_full"], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]) - 30.0)
        npt.assert_array_equal(self.unknowns["d_full"], 5.0)
        npt.assert_array_equal(self.unknowns["t_full"], 0.05)

    def testRefine3(self):
        mydis = tow.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        for k in self.unknowns["z_param"]:
            self.assertIn(k, self.unknowns["z_full"])

    def testRefineInterp(self):
        self.params["diameter"] = np.array([5.0, 5.0, 6.0, 7.0, 7.0])
        self.params["wall_thickness"] = 1e-2 * np.array([5.0, 5.0, 6.0, 7.0])
        mydis = tow.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns["z_param"], np.array([0.0, 1.0, 3.0, 6.0, 10.0]))
        npt.assert_array_equal(self.unknowns["z_full"], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]))
        npt.assert_array_equal(self.unknowns["d_full"], np.array([5.0, 5.0, 5.0, 5.5, 6.0, 6.5, 7.0, 7.0, 7.0]))
        npt.assert_array_equal(self.unknowns["t_full"], 1e-2 * np.array([5.0, 5.0, 5.0, 5.0, 6.0, 6.0, 7.0, 7.0]))


class TestMass(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params["d_full"] = 2.0 * 10.0 * myones
        self.params["t_full"] = 0.5 * np.ones((npts - 1,))
        self.params["z_full"] = np.linspace(0, 50.0, npts)
        self.params["rho"] = 5.0
        self.params["outfitting_factor"] = 1.5
        self.params["material_cost_rate"] = 1.5
        self.params["labor_cost_rate"] = 1.0
        self.params["painting_cost_rate"] = 10.0

        self.cm = tow.CylinderMass(nPoints=npts)

    def testRegular(self):
        # Straight column
        self.cm.compute(self.params, self.unknowns)

        expect = np.pi * (10.0 ** 2 - 9.5 ** 2) * 5.0 * 1.5 * (50.0 / (npts - 1))
        m = expect * (npts - 1)
        Iax = 0.5 * m * (10.0 ** 2 + 9.5 ** 2)
        Ix = (1 / 12.0) * m * (3 * (10.0 ** 2 + 9.5 ** 2) + 50 * 50) + m * 25 * 25  # parallel axis on last term
        z_avg, _ = nodal2sectional(self.params["z_full"])
        self.assertAlmostEqual(self.unknowns["mass"].sum(), m)
        npt.assert_almost_equal(self.unknowns["mass"], expect)
        npt.assert_almost_equal(self.unknowns["section_center_of_mass"], z_avg)
        self.assertAlmostEqual(self.unknowns["center_of_mass"], 25.0)
        npt.assert_almost_equal(self.unknowns["I_base"], [Ix, Ix, Iax, 0.0, 0.0, 0.0], decimal=5)

        """
    def testFrustum(self):
        # Frustum shell
        self.params['t_full'] = np.array([0.5, 0.4, 0.3])
        self.params['d_full'] = 2*np.array([10.0, 8.0, 6.0])
        self.wave.compute(self.params, self.unknowns)

        expect = np.pi/3.0*5.0*1.5*np.array([20.0, 30.0])*np.array([9.75*1.4+7.8*1.3, 7.8*1.1+5.85*1.0])
        m = expect*(npts-1)
        self.assertAlmostEqual(self.unknowns['mass'].sum(), m)
        npt.assert_almost_equal(self.unknowns['mass'].sum(), expect)
        """


class TestProps(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        # Store analysis options
        self.modeling_options = {}
        self.modeling_options["materials"] = {}
        self.modeling_options["materials"]["n_mat"] = 1

        self.modeling_options["flags"] = {}
        self.modeling_options["flags"]["monopile"] = False

        self.modeling_options["WISDEM"] = {}
        self.modeling_options["WISDEM"]["TowerSE"] = {}
        self.modeling_options["WISDEM"]["TowerSE"]["buckling_method"] = "eurocode"
        self.modeling_options["WISDEM"]["TowerSE"]["buckling_length"] = 30.0
        self.modeling_options["WISDEM"]["TowerSE"]["n_height_tower"] = 3
        self.modeling_options["WISDEM"]["TowerSE"]["n_layers_tower"] = 1
        self.modeling_options["WISDEM"]["TowerSE"]["n_height_monopile"] = 0
        self.modeling_options["WISDEM"]["TowerSE"]["n_layers_monopile"] = 0
        self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = 3
        self.modeling_options["WISDEM"]["TowerSE"]["wind"] = "PowerWind"
        self.modeling_options["WISDEM"]["TowerSE"]["nLC"] = 1

        self.modeling_options["WISDEM"]["TowerSE"]["soil_springs"] = False
        self.modeling_options["WISDEM"]["TowerSE"]["gravity_foundation"] = False

        self.modeling_options["WISDEM"]["TowerSE"]["gamma_f"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_m"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_n"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_b"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_fatigue"] = 1.0

        # Simplified the options available to the user
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"] = {}
        # self.modeling_options['TowerSE']['frame3dd']['DC']      = 80.0
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["shear"] = True
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["geom"] = True
        # self.modeling_options['TowerSE']['frame3dd']['dx']      = -1
        # self.modeling_options['TowerSE']['frame3dd']['nM']      = 6
        # self.modeling_options['TowerSE']['frame3dd']['Mmethod'] = 1
        # self.modeling_options['TowerSE']['frame3dd']['lump']    = 0
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["tol"] = 1e-9
        # self.modeling_options['TowerSE']['frame3dd']['shift']   = 0.0
        # self.modeling_options['TowerSE']['frame3dd']['add_gravity'] = True

    def testDiscYAML_Land_1Material(self):

        # Test land based, 1 material
        self.inputs["water_depth"] = 0.0
        self.inputs["tower_s"] = np.linspace(0, 1, 5)
        self.inputs["tower_layer_thickness"] = 0.25 * np.ones((1, 5))
        self.inputs["tower_foundation_height"] = 0.0
        self.inputs["tower_height"] = 1e2
        self.inputs["tower_outer_diameter_in"] = 8 * np.ones(5)
        self.inputs["tower_outfitting_factor"] = 1.1
        self.discrete_inputs["tower_layer_materials"] = ["steel"]
        self.inputs["monopile_s"] = np.empty(0)
        self.inputs["monopile_layer_thickness"] = np.empty((0, 0))
        self.inputs["monopile_foundation_height"] = 0.0
        self.inputs["monopile_height"] = 0.0
        self.inputs["monopile_outer_diameter_in"] = np.empty(0)
        self.inputs["monopile_outfitting_factor"] = 0.0
        self.discrete_inputs["monopile_layer_materials"] = [""]
        self.inputs["E_mat"] = 1e9 * np.ones((1, 3))
        self.inputs["G_mat"] = 1e8 * np.ones((1, 3))
        self.inputs["sigma_y_mat"] = np.array([1e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.ones((1, 3))
        self.inputs["wohler_exp_mat"] = np.array([1e1])
        self.inputs["wohler_A_mat"] = np.array([1e1])
        self.inputs["rho_mat"] = np.array([1e4])
        self.inputs["unit_cost_mat"] = np.array([1e1])
        self.discrete_inputs["material_names"] = ["steel"]
        myobj = tow.DiscretizationYAML(
            n_height_tower=5, n_height_monopile=0, n_layers_tower=1, n_layers_monopile=0, n_mat=1
        )
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(self.outputs["tower_section_height"], 25.0 * np.ones(4))
        npt.assert_equal(self.outputs["tower_outer_diameter"], self.inputs["tower_outer_diameter_in"])
        npt.assert_equal(self.outputs["tower_wall_thickness"], 0.25 * np.ones(4))
        npt.assert_equal(self.outputs["outfitting_factor"], 1.1 * np.ones(4))
        npt.assert_equal(self.outputs["E"], 1e9 * np.ones(4))
        npt.assert_equal(self.outputs["G"], 1e8 * np.ones(4))
        npt.assert_equal(self.outputs["sigma_y"], 1e7 * np.ones(4))
        npt.assert_equal(self.outputs["sigma_ult"], 1e7 * np.ones(4))
        npt.assert_equal(self.outputs["wohler_exp"], 1e1 * np.ones(4))
        npt.assert_equal(self.outputs["wohler_A"], 1e1 * np.ones(4))
        npt.assert_equal(self.outputs["rho"], 1e4 * np.ones(4))
        npt.assert_equal(self.outputs["unit_cost"], 1e1 * np.ones(4))
        npt.assert_equal(self.outputs["z_start"], 0.0)
        npt.assert_equal(self.outputs["transition_piece_height"], 0.0)
        npt.assert_equal(self.outputs["suctionpile_depth"], 0.0)

    def testDiscYAML_Land_2Materials(self):
        # Test land based, 2 materials
        self.inputs["water_depth"] = 0.0
        self.inputs["tower_s"] = np.linspace(0, 1, 5)
        self.inputs["tower_layer_thickness"] = np.array([[0.2, 0.2, 0.2, 0.0, 0.0], [0.0, 0.0, 0.0, 0.1, 0.1]])
        self.inputs["tower_foundation_height"] = 0.0
        self.inputs["tower_height"] = 1e2
        self.inputs["tower_outer_diameter_in"] = 8 * np.ones(5)
        self.inputs["tower_outfitting_factor"] = 1.1
        self.discrete_inputs["tower_layer_materials"] = ["steel", "other"]
        self.inputs["monopile_s"] = np.empty(0)
        self.inputs["monopile_layer_thickness"] = np.empty((0, 0))
        self.inputs["monopile_foundation_height"] = 0.0
        self.inputs["monopile_height"] = 0.0
        self.inputs["monopile_outer_diameter_in"] = np.empty(0)
        self.inputs["monopile_outfitting_factor"] = 0.0
        self.discrete_inputs["monopile_layer_materials"] = [""]
        self.inputs["E_mat"] = 1e9 * np.vstack((np.ones((1, 3)), 2 * np.ones((1, 3))))
        self.inputs["G_mat"] = 1e8 * np.vstack((np.ones((1, 3)), 2 * np.ones((1, 3))))
        self.inputs["sigma_y_mat"] = np.array([1e7, 2e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.vstack((np.ones((1, 3)), 2 * np.ones((1, 3))))
        self.inputs["wohler_exp_mat"] = np.array([1e1, 2e1])
        self.inputs["wohler_A_mat"] = np.array([1e1, 2e1])
        self.inputs["rho_mat"] = np.array([1e4, 2e4])
        self.inputs["unit_cost_mat"] = np.array([1e1, 2e1])
        self.discrete_inputs["material_names"] = ["steel", "other"]
        myobj = tow.DiscretizationYAML(
            n_height_tower=5, n_height_monopile=0, n_layers_tower=1, n_layers_monopile=0, n_mat=2
        )
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        # Define mixtures
        v = np.r_[np.mean([0.2, 0]), np.mean([0.1, 0.0])]
        vv = v / v.sum()
        x = np.r_[1, 2]
        xx1 = np.sum(x * vv)  # Mass weighted
        xx2 = 0.5 * np.sum(vv * x) + 0.5 / np.sum(vv / x)  # Volumetric method
        xx3 = np.sum(x * x * vv) / xx1  # Mass-cost weighted
        npt.assert_equal(self.outputs["tower_section_height"], 25.0 * np.ones(4))
        npt.assert_equal(self.outputs["tower_outer_diameter"], self.inputs["tower_outer_diameter_in"])
        npt.assert_almost_equal(self.outputs["tower_wall_thickness"], np.array([0.2, 0.2, v.sum(), 0.1]))
        npt.assert_equal(self.outputs["outfitting_factor"], 1.1 * np.ones(4))
        npt.assert_almost_equal(self.outputs["E"], 1e9 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(self.outputs["G"], 1e8 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(self.outputs["sigma_y"], 1e7 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(self.outputs["sigma_ult"], 1e7 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(self.outputs["rho"], 1e4 * np.array([1, 1, xx1, 2]))
        npt.assert_almost_equal(self.outputs["wohler_exp"], 1e1 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(self.outputs["wohler_A"], 1e1 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(self.outputs["unit_cost"], 1e1 * np.array([1, 1, xx3, 2]))
        npt.assert_equal(self.outputs["z_start"], 0.0)
        npt.assert_equal(self.outputs["transition_piece_height"], 0.0)
        npt.assert_equal(self.outputs["suctionpile_depth"], 0.0)

    def testDiscYAML_Monopile_1Material(self):
        self.inputs["water_depth"] = 30.0
        self.inputs["tower_s"] = np.linspace(0, 1, 5)
        self.inputs["tower_layer_thickness"] = 0.25 * np.ones((1, 5))
        self.inputs["tower_height"] = 1e2
        self.inputs["tower_foundation_height"] = 10.0
        self.inputs["tower_outer_diameter_in"] = 8 * np.ones(5)
        self.inputs["tower_outfitting_factor"] = 1.1
        self.discrete_inputs["tower_layer_materials"] = ["steel"]
        self.inputs["monopile_s"] = np.linspace(0, 1, 5)
        self.inputs["monopile_layer_thickness"] = 0.5 * np.ones((1, 5))
        self.inputs["monopile_foundation_height"] = -40.0
        self.inputs["monopile_height"] = 50.0
        self.inputs["monopile_outer_diameter_in"] = 10 * np.ones(5)
        self.inputs["monopile_outer_diameter_in"][-1] = 8
        self.inputs["monopile_outfitting_factor"] = 1.2
        self.discrete_inputs["monopile_layer_materials"] = ["steel"]
        self.inputs["E_mat"] = 1e9 * np.ones((1, 3))
        self.inputs["G_mat"] = 1e8 * np.ones((1, 3))
        self.inputs["sigma_y_mat"] = np.array([1e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.ones((1, 3))
        self.inputs["wohler_exp_mat"] = np.array([1e1])
        self.inputs["wohler_A_mat"] = np.array([1e1])
        self.inputs["rho_mat"] = np.array([1e4])
        self.inputs["unit_cost_mat"] = np.array([1e1])
        self.discrete_inputs["material_names"] = ["steel"]
        myobj = tow.DiscretizationYAML(
            n_height_tower=5, n_height_monopile=5, n_layers_tower=1, n_layers_monopile=1, n_mat=1
        )
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(self.outputs["tower_section_height"], np.r_[10.0, 15.0, 12.5 * np.ones(2), 25 * np.ones(4)])
        npt.assert_equal(self.outputs["tower_outer_diameter"], np.r_[10 * np.ones(4), 8 * np.ones(5)])
        npt.assert_equal(self.outputs["tower_wall_thickness"], np.r_[0.5 * np.ones(4), 0.25 * np.ones(4)])
        npt.assert_equal(self.outputs["outfitting_factor"], np.r_[1.2 * np.ones(4), 1.1 * np.ones(4)])
        npt.assert_equal(self.outputs["E"], 1e9 * np.ones(8))
        npt.assert_equal(self.outputs["G"], 1e8 * np.ones(8))
        npt.assert_equal(self.outputs["sigma_y"], 1e7 * np.ones(8))
        npt.assert_equal(self.outputs["sigma_ult"], 1e7 * np.ones(8))
        npt.assert_equal(self.outputs["wohler_exp"], 1e1 * np.ones(8))
        npt.assert_equal(self.outputs["wohler_A"], 1e1 * np.ones(8))
        npt.assert_equal(self.outputs["rho"], 1e4 * np.ones(8))
        npt.assert_equal(self.outputs["unit_cost"], 1e1 * np.ones(8))
        npt.assert_equal(self.outputs["z_start"], -40.0)
        npt.assert_equal(self.outputs["transition_piece_height"], 10.0)
        npt.assert_equal(self.outputs["suctionpile_depth"], 10.0)

    def testDiscYAML_Monopile_PileShort(self):
        self.inputs["water_depth"] = 60.0
        self.inputs["tower_s"] = np.linspace(0, 1, 5)
        self.inputs["tower_layer_thickness"] = 0.25 * np.ones((1, 5))
        self.inputs["tower_height"] = 1e2
        self.inputs["tower_foundation_height"] = 10.0
        self.inputs["tower_outer_diameter_in"] = 8 * np.ones(5)
        self.inputs["tower_outfitting_factor"] = 1.1
        self.discrete_inputs["tower_layer_materials"] = ["steel"]
        self.inputs["monopile_s"] = np.linspace(0, 1, 5)
        self.inputs["monopile_layer_thickness"] = 0.5 * np.ones((1, 5))
        self.inputs["monopile_foundation_height"] = -40.0
        self.inputs["monopile_height"] = 50.0
        self.inputs["monopile_outer_diameter_in"] = 10 * np.ones(5)
        self.inputs["monopile_outer_diameter_in"][-1] = 8
        self.inputs["monopile_outfitting_factor"] = 1.2
        self.discrete_inputs["monopile_layer_materials"] = ["steel"]
        self.inputs["E_mat"] = 1e9 * np.ones((1, 3))
        self.inputs["G_mat"] = 1e8 * np.ones((1, 3))
        self.inputs["sigma_y_mat"] = np.array([1e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.ones((1, 3))
        self.inputs["wohler_exp_mat"] = np.array([1e1])
        self.inputs["wohler_A_mat"] = np.array([1e1])
        self.inputs["rho_mat"] = np.array([1e4])
        self.inputs["unit_cost_mat"] = np.array([1e1])
        self.discrete_inputs["material_names"] = ["steel"]
        myobj = tow.DiscretizationYAML(
            n_height_tower=5, n_height_monopile=5, n_layers_tower=1, n_layers_monopile=1, n_mat=1
        )
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(self.outputs["tower_section_height"], np.r_[12.5 * np.ones(4), 25 * np.ones(4)])
        npt.assert_equal(self.outputs["tower_outer_diameter"], np.r_[10 * np.ones(4), 8 * np.ones(5)])
        npt.assert_equal(self.outputs["tower_wall_thickness"], np.r_[0.5 * np.ones(4), 0.25 * np.ones(4)])
        npt.assert_equal(self.outputs["outfitting_factor"], np.r_[1.2 * np.ones(4), 1.1 * np.ones(4)])
        npt.assert_equal(self.outputs["E"], 1e9 * np.ones(8))
        npt.assert_equal(self.outputs["G"], 1e8 * np.ones(8))
        npt.assert_equal(self.outputs["sigma_y"], 1e7 * np.ones(8))
        npt.assert_equal(self.outputs["sigma_ult"], 1e7 * np.ones(8))
        npt.assert_equal(self.outputs["wohler_exp"], 1e1 * np.ones(8))
        npt.assert_equal(self.outputs["wohler_A"], 1e1 * np.ones(8))
        npt.assert_equal(self.outputs["rho"], 1e4 * np.ones(8))
        npt.assert_equal(self.outputs["unit_cost"], 1e1 * np.ones(8))
        npt.assert_equal(self.outputs["z_start"], -40.0)
        npt.assert_equal(self.outputs["transition_piece_height"], 10.0)
        npt.assert_equal(self.outputs["suctionpile_depth"], -20.0)

    def testDiscYAML_Monopile_RedoPileNodes(self):
        self.inputs["water_depth"] = 30.0
        self.inputs["tower_s"] = np.linspace(0, 1, 5)
        self.inputs["tower_layer_thickness"] = 0.25 * np.ones((1, 5))
        self.inputs["tower_height"] = 1e2
        self.inputs["tower_foundation_height"] = 10.0
        self.inputs["tower_outer_diameter_in"] = 8 * np.ones(5)
        self.inputs["tower_outfitting_factor"] = 1.1
        self.discrete_inputs["tower_layer_materials"] = ["steel"]
        self.inputs["monopile_s"] = np.linspace(0, 1, 20)
        self.inputs["monopile_layer_thickness"] = 0.5 * np.ones((1, 20))
        self.inputs["monopile_foundation_height"] = -40.0
        self.inputs["monopile_height"] = 50.0
        self.inputs["monopile_outer_diameter_in"] = 10 * np.ones(20)
        self.inputs["monopile_outer_diameter_in"][-1] = 8
        self.inputs["monopile_outfitting_factor"] = 1.2
        self.discrete_inputs["monopile_layer_materials"] = ["steel"]
        self.inputs["E_mat"] = 1e9 * np.ones((1, 3))
        self.inputs["G_mat"] = 1e8 * np.ones((1, 3))
        self.inputs["sigma_y_mat"] = np.array([1e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.ones((1, 3))
        self.inputs["wohler_exp_mat"] = np.array([1e1])
        self.inputs["wohler_A_mat"] = np.array([1e1])
        self.inputs["rho_mat"] = np.array([1e4])
        self.inputs["unit_cost_mat"] = np.array([1e1])
        self.discrete_inputs["material_names"] = ["steel"]
        myobj = tow.DiscretizationYAML(
            n_height_tower=5, n_height_monopile=20, n_layers_tower=1, n_layers_monopile=1, n_mat=1
        )
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(self.outputs["tower_section_height"][0], 10.0)
        # npt.assert_almost_equal(self.outputs["tower_section_height"][1:4], (2.63157895*4-10)/4)
        npt.assert_almost_equal(
            self.outputs["tower_section_height"][4:], np.r_[2.63157895 * np.ones(15), 25 * np.ones(4)]
        )
        npt.assert_equal(self.outputs["tower_outer_diameter"], np.r_[10 * np.ones(19), 8 * np.ones(5)])
        npt.assert_equal(self.outputs["tower_wall_thickness"], np.r_[0.5 * np.ones(19), 0.25 * np.ones(4)])
        npt.assert_equal(self.outputs["outfitting_factor"], np.r_[1.2 * np.ones(19), 1.1 * np.ones(4)])
        npt.assert_equal(self.outputs["E"], 1e9 * np.ones(23))
        npt.assert_equal(self.outputs["G"], 1e8 * np.ones(23))
        npt.assert_equal(self.outputs["sigma_y"], 1e7 * np.ones(23))
        npt.assert_equal(self.outputs["sigma_ult"], 1e7 * np.ones(23))
        npt.assert_equal(self.outputs["wohler_exp"], 1e1 * np.ones(23))
        npt.assert_equal(self.outputs["wohler_A"], 1e1 * np.ones(23))
        npt.assert_equal(self.outputs["rho"], 1e4 * np.ones(23))
        npt.assert_equal(self.outputs["unit_cost"], 1e1 * np.ones(23))
        npt.assert_equal(self.outputs["z_start"], -40.0)
        npt.assert_equal(self.outputs["transition_piece_height"], 10.0)
        npt.assert_equal(self.outputs["suctionpile_depth"], 10.0)

    def testDiscYAML_Monopile_DifferentMaterials(self):
        self.inputs["water_depth"] = 30.0
        self.inputs["tower_s"] = np.linspace(0, 1, 5)
        self.inputs["tower_layer_thickness"] = 0.25 * np.ones((1, 5))
        self.inputs["tower_foundation_height"] = 10.0
        self.inputs["tower_height"] = 1e2
        self.inputs["tower_outer_diameter_in"] = 8 * np.ones(5)
        self.inputs["tower_outfitting_factor"] = 1.1
        self.discrete_inputs["tower_layer_materials"] = ["steel"]
        self.inputs["monopile_s"] = np.linspace(0, 1, 5)
        self.inputs["monopile_layer_thickness"] = 0.5 * np.ones((1, 5))
        self.inputs["monopile_foundation_height"] = -40.0
        self.inputs["monopile_height"] = 50.0
        self.inputs["monopile_outer_diameter_in"] = 10 * np.ones(5)
        self.inputs["monopile_outer_diameter_in"][-1] = 8
        self.inputs["monopile_outfitting_factor"] = 1.2
        self.discrete_inputs["monopile_layer_materials"] = ["other"]
        self.inputs["E_mat"] = 1e9 * np.vstack((np.ones((1, 3)), 2 * np.ones((1, 3))))
        self.inputs["G_mat"] = 1e8 * np.vstack((np.ones((1, 3)), 2 * np.ones((1, 3))))
        self.inputs["sigma_y_mat"] = np.array([1e7, 2e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.vstack((np.ones((1, 3)), 2 * np.ones((1, 3))))
        self.inputs["wohler_exp_mat"] = np.array([1e1, 2e1])
        self.inputs["wohler_A_mat"] = np.array([1e1, 2e1])
        self.inputs["rho_mat"] = np.array([1e4, 2e4])
        self.inputs["unit_cost_mat"] = np.array([1e1, 2e1])
        self.discrete_inputs["material_names"] = ["steel", "other"]
        myobj = tow.DiscretizationYAML(
            n_height_tower=5, n_height_monopile=5, n_layers_tower=1, n_layers_monopile=1, n_mat=2
        )
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(self.outputs["tower_section_height"], np.r_[10.0, 15.0, 12.5 * np.ones(2), 25 * np.ones(4)])
        npt.assert_equal(self.outputs["tower_outer_diameter"], np.r_[10 * np.ones(4), 8 * np.ones(5)])
        npt.assert_equal(self.outputs["tower_wall_thickness"], np.r_[0.5 * np.ones(4), 0.25 * np.ones(4)])
        npt.assert_equal(self.outputs["outfitting_factor"], np.r_[1.2 * np.ones(4), 1.1 * np.ones(4)])
        npt.assert_equal(self.outputs["E"], 1e9 * np.r_[2 * np.ones(4), np.ones(4)])
        npt.assert_equal(self.outputs["G"], 1e8 * np.r_[2 * np.ones(4), np.ones(4)])
        npt.assert_equal(self.outputs["sigma_y"], 1e7 * np.r_[2 * np.ones(4), np.ones(4)])
        npt.assert_equal(self.outputs["sigma_ult"], 1e7 * np.r_[2 * np.ones(4), np.ones(4)])
        npt.assert_equal(self.outputs["wohler_exp"], 1e1 * np.r_[2 * np.ones(4), np.ones(4)])
        npt.assert_equal(self.outputs["wohler_A"], 1e1 * np.r_[2 * np.ones(4), np.ones(4)])
        npt.assert_equal(self.outputs["rho"], 1e4 * np.r_[2 * np.ones(4), np.ones(4)])
        npt.assert_equal(self.outputs["unit_cost"], 1e1 * np.r_[2 * np.ones(4), np.ones(4)])
        npt.assert_equal(self.outputs["z_start"], -40.0)
        npt.assert_equal(self.outputs["transition_piece_height"], 10.0)
        npt.assert_equal(self.outputs["suctionpile_depth"], 10.0)

    def testDiscYAML_Bad_Inputs(self):
        self.inputs["water_depth"] = 0.0
        self.inputs["tower_s"] = np.linspace(0, 1, 5)
        self.inputs["tower_layer_thickness"] = 0.25 * np.ones((1, 5))
        self.inputs["tower_foundation_height"] = 0.0
        self.inputs["tower_height"] = 1e2
        self.inputs["tower_outer_diameter_in"] = 8 * np.ones(5)
        self.inputs["tower_outfitting_factor"] = 1.1
        self.discrete_inputs["tower_layer_materials"] = ["steel"]
        self.inputs["monopile_s"] = np.empty(0)
        self.inputs["monopile_layer_thickness"] = np.empty((0, 0))
        self.inputs["monopile_foundation_height"] = 0.0
        self.inputs["monopile_height"] = 0.0
        self.inputs["monopile_outer_diameter_in"] = np.empty(0)
        self.inputs["monopile_outfitting_factor"] = 0.0
        self.discrete_inputs["monopile_layer_materials"] = [""]
        self.inputs["E_mat"] = 1e9 * np.ones((1, 3))
        self.inputs["G_mat"] = 1e8 * np.ones((1, 3))
        self.inputs["sigma_y_mat"] = np.array([1e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.ones((1, 3))
        self.inputs["wohler_exp_mat"] = np.array([1e1])
        self.inputs["wohler_A_mat"] = np.array([1e1])
        self.inputs["rho_mat"] = np.array([1e4])
        self.inputs["unit_cost_mat"] = np.array([1e1])
        self.discrete_inputs["material_names"] = ["steel"]
        myobj = tow.DiscretizationYAML(
            n_height_tower=5, n_height_monopile=0, n_layers_tower=1, n_layers_monopile=0, n_mat=1
        )

        try:
            self.inputs["tower_layer_thickness"][0, -2:] = 0.0
            myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
            self.assertTrue(False)  # Shouldn't get here
        except ValueError:
            self.assertTrue(True)

        try:
            self.inputs["tower_layer_thickness"][0, -2:] = 0.25
            self.inputs["tower_outer_diameter_in"][-1] = -1.0
            myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
            self.assertTrue(False)  # Shouldn't get here
        except ValueError:
            self.assertTrue(True)

        try:
            self.inputs["tower_layer_thickness"][0, -2:] = 0.25
            self.inputs["tower_outer_diameter_in"][-1] = 8.0
            self.inputs["tower_s"][-1] = self.inputs["tower_s"][-2]
            myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
            self.assertTrue(False)  # Shouldn't get here
        except ValueError:
            self.assertTrue(True)

    def testTowerDisc(self):
        # Test Land
        self.inputs["hub_height"] = 100.0
        self.inputs["z_param"] = np.array([0.0, 40.0, 80.0])
        self.inputs["z_full"] = np.linspace(0.0, 80.0, 7)
        self.inputs["tower_outer_diameter"] = 10 * np.ones(3)
        self.inputs["rho"] = 1e3 * np.ones(2)
        self.inputs["outfitting_factor"] = 1.1 * np.ones(2)
        self.inputs["unit_cost"] = 5.0 * np.ones(2)
        self.inputs["E"] = 6.0 * np.ones(2)
        self.inputs["G"] = 7.0 * np.ones(2)
        self.inputs["sigma_y"] = 8.0 * np.ones(2)
        self.inputs["sigma_ult"] = 8.0 * np.ones(2)
        self.inputs["wohler_exp"] = 1e1 * np.ones(2)
        self.inputs["wohler_A"] = 1e1 * np.ones(2)
        self.inputs["Asx"] = 8.0 * np.ones(6)
        self.inputs["Asy"] = 8.0 * np.ones(6)
        self.inputs["Az"] = 9.0 * np.ones(6)
        self.inputs["Jz"] = 10.0 * np.ones(6)
        self.inputs["Ixx"] = 11.0 * np.ones(6)
        self.inputs["Iyy"] = 11.0 * np.ones(6)
        myobj = tow.TowerDiscretization(n_height=3)
        myobj.compute(self.inputs, self.outputs)
        self.assertEqual(self.outputs["height_constraint"], 20.0)
        npt.assert_equal(self.outputs["rho_full"], self.inputs["rho"][0] * np.ones(6))
        npt.assert_equal(self.outputs["E_full"], self.inputs["E"][0] * np.ones(6))
        npt.assert_equal(self.outputs["G_full"], self.inputs["G"][0] * np.ones(6))
        npt.assert_equal(self.outputs["sigma_y_full"], self.inputs["sigma_y"][0] * np.ones(6))
        # npt.assert_equal(self.outputs["sigma_ult_full"], self.inputs["sigma_ult"][0] * np.ones(6))
        npt.assert_equal(self.outputs["unit_cost_full"], self.inputs["unit_cost"][0] * np.ones(6))

        nout = 2
        npt.assert_almost_equal(self.outputs["sec_loc"], np.linspace(0, 1, nout))
        # npt.assert_equal(self.outputs["str_tw"], np.zeros(nout))
        # npt.assert_equal(self.outputs["tw_iner"], np.zeros(nout))
        of = self.inputs["outfitting_factor"]
        npt.assert_equal(self.outputs["mass_den"], 1e3 * 9 * np.ones(nout) * of)
        npt.assert_equal(self.outputs["foreaft_iner"], 1e3 * 11 * np.ones(nout))
        npt.assert_equal(self.outputs["sideside_iner"], 1e3 * 11 * np.ones(nout))
        npt.assert_equal(self.outputs["foreaft_stff"], 6 * 11 * np.ones(nout))
        npt.assert_equal(self.outputs["sideside_stff"], 6 * 11 * np.ones(nout))
        npt.assert_equal(self.outputs["tor_stff"], 7 * 10 * np.ones(nout))
        npt.assert_equal(self.outputs["axial_stff"], 6 * 9 * np.ones(nout))
        # npt.assert_equal(self.outputs["cg_offst"], np.zeros(nout))
        # npt.assert_equal(self.outputs["sc_offst"], np.zeros(nout))
        # npt.assert_equal(self.outputs["tc_offst"], np.zeros(nout))

        r = 0.5 * 10.0
        npt.assert_almost_equal(self.outputs["axial_load2stress"][:, 2], 1.0 / 9.0)
        npt.assert_equal(self.outputs["axial_load2stress"][:, :2], 0.0)
        npt.assert_equal(self.outputs["axial_load2stress"][:, 5], 0.0)
        npt.assert_almost_equal(self.outputs["axial_load2stress"][:, 3:4], r / 11.0)
        npt.assert_almost_equal(self.outputs["shear_load2stress"][:, :2], 1 / 8.0)
        npt.assert_almost_equal(self.outputs["shear_load2stress"][:, 2:4], 0.0)
        npt.assert_almost_equal(self.outputs["shear_load2stress"][:, 5], r / 10.0)

    def testTowerMass(self):

        self.inputs["z_full"] = np.array([-50.0, -30, 0.0, 40.0, 80.0])
        self.inputs["d_full"] = 10 * np.ones(5)
        self.inputs["cylinder_mass"] = 1e3 * np.ones(4)
        self.inputs["cylinder_cost"] = 1e5
        self.inputs["cylinder_center_of_mass"] = 10.0
        self.inputs["cylinder_section_center_of_mass"] = self.inputs["z_full"][:-1] + 0.5 * np.diff(
            self.inputs["z_full"]
        )
        self.inputs["cylinder_I_base"] = 1e4 * np.r_[np.ones(3), np.zeros(3)]
        self.inputs["transition_piece_height"] = 20.0
        self.inputs["transition_piece_mass"] = 1e2
        self.inputs["transition_piece_cost"] = 1e3
        self.inputs["gravity_foundation_mass"] = 1e2

        myobj = tow.TowerMass(n_height=5)
        myobj.compute(self.inputs, self.outputs)

        self.assertEqual(self.outputs["structural_cost"], self.inputs["cylinder_cost"] + 1e3)
        npt.assert_equal(self.outputs["tower_I_base"], self.inputs["cylinder_I_base"])
        self.assertEqual(
            self.outputs["tower_center_of_mass"], (4 * 1e3 * 10.0 + 1e2 * 20.0 + 1e2 * -50.0) / (4 * 1e3 + 2e2)
        )
        npt.assert_equal(self.outputs["tower_section_center_of_mass"], self.inputs["cylinder_section_center_of_mass"])
        self.assertEqual(self.outputs["monopile_mass"], 1e3 * 2.5 + 2 * 1e2)
        self.assertEqual(self.outputs["monopile_cost"], self.inputs["cylinder_cost"] * 2.5 / 4.0 + 1e3)
        self.assertEqual(self.outputs["tower_mass"], 1e3 * (4 - 2.5))
        self.assertEqual(self.outputs["tower_cost"], self.inputs["cylinder_cost"] * 1.5 / 4.0)
        npt.assert_equal(self.outputs["transition_piece_I"], 1e2 * 25 * np.r_[0.5, 0.5, 1.0, np.zeros(3)])
        npt.assert_equal(self.outputs["gravity_foundation_I"], 0.5 * 1e2 * 25 * np.r_[0.5, 0.5, 1.0, np.zeros(3)])


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestDiscretization))
    suite.addTest(unittest.makeSuite(TestMass))
    suite.addTest(unittest.makeSuite(TestProps))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



"""`MooringSystemDesign` and related functionality."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from math import sqrt

from wisdem.orbit.phases.design import DesignPhase


class MooringSystemDesign(DesignPhase):
    """Mooring System and Anchor Design."""

    expected_config = {
        "site": {"depth": "float"},
        "turbine": {"turbine_rating": "int | float"},
        "plant": {"num_turbines": "int"},
        "mooring_system_design": {
            "num_lines": "int | float (optional, default: 4)",
            "anchor_type": "str (optional, default: 'Suction Pile')",
            "mooring_line_cost_rate": "int | float (optional)",
            "drag_embedment_fixed_length": "int | float (optional, default: .5km)",
        },
    }

    output_config = {
        "mooring_system": {
            "num_lines": "int",
            "line_diam": "m, float",
            "line_mass": "t",
            "line_cost": "USD",
            "line_length": "m",
            "anchor_mass": "t",
            "anchor_type": "str",
            "anchor_cost": "USD",
        }
    }

    def __init__(self, config, **kwargs):
        """
        Creates an instance of MooringSystemDesign.

        Parameters
        ----------
        config : dict
        """

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)
        self.num_turbines = self.config["plant"]["num_turbines"]

        self._design = self.config.get("mooring_system_design", {})
        self.num_lines = self._design.get("num_lines", 4)
        self.anchor_type = self._design.get("anchor_type", "Suction Pile")

        self._outputs = {}

    def run(self):
        """
        Main run function.
        """

        self.determine_mooring_line()
        self.calculate_breaking_load()
        self.calculate_line_length_mass()
        self.calculate_anchor_mass_cost()

        self._outputs["mooring_system"] = {**self.design_result}

    def determine_mooring_line(self):
        """
        Returns the diameter of the mooring lines based on the turbine rating.
        """

        tr = self.config["turbine"]["turbine_rating"]
        fit = -0.0004 * (tr ** 2) + 0.0132 * tr + 0.0536

        if fit <= 0.09:
            self.line_diam = 0.09
            self.line_mass_per_m = 0.161
            self.line_cost_rate = 399.0

        elif fit <= 0.12:
            self.line_diam = 0.12
            self.line_mass_per_m = 0.288
            self.line_cost_rate = 721.0

        else:
            self.line_diam = 0.15
            self.line_mass_per_m = 0.450
            self.line_cost_rate = 1088.0

    def calculate_breaking_load(self):
        """
        Returns the mooring line breaking load.
        """

        self.breaking_load = 419449 * (self.line_diam ** 2) + 93415 * self.line_diam - 3577.9

    def calculate_line_length_mass(self):
        """
        Returns the mooring line length and mass.
        """

        if self.anchor_type == "Drag Embedment":
            fixed = self._design.get("drag_embedment_fixed_length", 0.5)

        else:
            fixed = 0

        depth = self.config["site"]["depth"]
        self.line_length = 0.0002 * (depth ** 2) + 1.264 * depth + 47.776 + fixed

        self.line_mass = self.line_length * self.line_mass_per_m

    def calculate_anchor_mass_cost(self):
        """
        Returns the mass and cost of anchors.

        TODO: Anchor masses are rough estimates based on initial literature
        review. Should be revised when this module is overhauled in the future.
        """

        if self.anchor_type == "Drag Embedment":
            self.anchor_mass = 20
            self.anchor_cost = self.breaking_load / 9.81 / 20.0 * 2000.0

        else:
            self.anchor_mass = 50
            self.anchor_cost = sqrt(self.breaking_load / 9.81 / 1250) * 150000

    @property
    def line_cost(self):
        """Returns cost of one line mooring line."""

        return self.line_length * self.line_cost_rate

    @property
    def total_cost(self):
        """Returns the total cost of the mooring system."""

        return self.num_lines * self.num_turbines * (self.anchor_cost + self.line_length * self.line_cost_rate)

    @property
    def detailed_output(self):
        """Returns detailed phase information."""

        return {
            "num_lines": self.num_lines,
            "line_diam": self.line_diam,
            "line_mass": self.line_mass,
            "line_length": self.line_length,
            "line_cost": self.line_cost,
            "anchor_type": self.anchor_type,
            "anchor_mass": self.anchor_mass,
            "anchor_cost": self.anchor_cost,
            "system_cost": self.total_cost,
        }

    @property
    def design_result(self):
        """Returns the results of the design phase."""

        return {"mooring_system": self.detailed_output}



import openmdao.api as om
from wisdem.floatingse.member import Member
from wisdem.floatingse.mooring import Mooring
from wisdem.floatingse.constraints import FloatingConstraints
from wisdem.floatingse.floating_frame import FloatingFrame


class FloatingSE(om.Group):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]

        # self.set_input_defaults("mooring_type", "chain")
        # self.set_input_defaults("anchor_type", "SUCTIONPILE")
        # self.set_input_defaults("loading", "hydrostatic")
        # self.set_input_defaults("wave_period_range_low", 2.0, units="s")
        # self.set_input_defaults("wave_period_range_high", 20.0, units="s")
        # self.set_input_defaults("cd_usr", -1.0)
        # self.set_input_defaults("zref", 100.0)
        # self.set_input_defaults("number_of_offset_columns", 0)
        # self.set_input_defaults("material_names", ["steel"])

        n_member = opt["floating"]["members"]["n_members"]
        mem_prom = [
            "E_mat",
            "G_mat",
            "sigma_y_mat",
            "sigma_ult_mat",
            "wohler_exp_mat",
            "wohler_A_mat",
            "rho_mat",
            "rho_water",
            "unit_cost_mat",
            "material_names",
            "painting_cost_rate",
            "labor_cost_rate",
        ]
        mem_prom += [
            "Uref",
            "zref",
            "z0",
            "shearExp",
            "cd_usr",
            "cm",
            "beta_wind",
            "rho_air",
            "mu_air",
            "beta_wave",
            "mu_water",
            "Uc",
            "Hsig_wave",
            "Tsig_wave",
            "water_depth",
        ]
        for k in range(n_member):
            self.add_subsystem(
                f"member{k}",
                Member(column_options=opt["floating"]["members"], idx=k, n_mat=opt["materials"]["n_mat"]),
                promotes=mem_prom,
            )

        # Next run MapMooring
        self.add_subsystem(
            "mm", Mooring(options=opt["mooring"], gamma=opt["WISDEM"]["FloatingSE"]["gamma_f"]), promotes=["*"]
        )

        # Add in the connecting truss
        self.add_subsystem("load", FloatingFrame(modeling_options=opt), promotes=["*"])

        # Evaluate system constraints
        self.add_subsystem("cons", FloatingConstraints(modeling_options=opt), promotes=["*"])

        # Connect all input variables from all models
        mem_vars = [
            "nodes_xyz",
            "nodes_r",
            "section_D",
            "section_t",
            "section_A",
            "section_Asx",
            "section_Asy",
            "section_Ixx",
            "section_Iyy",
            "section_Izz",
            "section_rho",
            "section_E",
            "section_G",
            "section_sigma_y",
            "idx_cb",
            "variable_ballast_capacity",
            "variable_ballast_Vpts",
            "variable_ballast_spts",
            "constr_ballast_capacity",
            "buoyancy_force",
            "displacement",
            "center_of_buoyancy",
            "center_of_mass",
            "ballast_mass",
            "total_mass",
            "total_cost",
            "I_total",
            "Awater",
            "Iwater",
            "added_mass",
            "waterline_centroid",
            "Px",
            "Py",
            "Pz",
            "qdyn",
        ]
        for k in range(n_member):
            for var in mem_vars:
                self.connect(f"member{k}." + var, f"member{k}:" + var)



import unittest

import numpy as np
import numpy.testing as npt
import wisdem.towerse.tower_struct as tow
from wisdem.towerse import RIGID


class TestStruct(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        # Store analysis options
        self.modeling_options = {}
        self.modeling_options["materials"] = {}
        self.modeling_options["materials"]["n_mat"] = 1

        self.modeling_options["flags"] = {}
        self.modeling_options["flags"]["monopile"] = False

        self.modeling_options["WISDEM"] = {}
        self.modeling_options["WISDEM"]["TowerSE"] = {}
        self.modeling_options["WISDEM"]["TowerSE"]["buckling_method"] = "eurocode"
        self.modeling_options["WISDEM"]["TowerSE"]["buckling_length"] = 30.0
        self.modeling_options["WISDEM"]["TowerSE"]["n_height_tower"] = 3
        self.modeling_options["WISDEM"]["TowerSE"]["n_layers_tower"] = 1
        self.modeling_options["WISDEM"]["TowerSE"]["n_height_monopile"] = 0
        self.modeling_options["WISDEM"]["TowerSE"]["n_layers_monopile"] = 0
        self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = 3
        self.modeling_options["WISDEM"]["TowerSE"]["n_refine"] = 3
        self.modeling_options["WISDEM"]["TowerSE"]["wind"] = "PowerWind"
        self.modeling_options["WISDEM"]["TowerSE"]["nLC"] = 1

        self.modeling_options["WISDEM"]["TowerSE"]["soil_springs"] = False
        self.modeling_options["WISDEM"]["TowerSE"]["gravity_foundation"] = False

        self.modeling_options["WISDEM"]["TowerSE"]["gamma_f"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_m"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_n"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_b"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_fatigue"] = 1.0

        # Simplified the options available to the user
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"] = {}
        # self.modeling_options['TowerSE']['frame3dd']['DC']      = 80.0
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["shear"] = True
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["geom"] = True
        # self.modeling_options['TowerSE']['frame3dd']['dx']      = -1
        # self.modeling_options['TowerSE']['frame3dd']['nM']      = 6
        # self.modeling_options['TowerSE']['frame3dd']['Mmethod'] = 1
        # self.modeling_options['TowerSE']['frame3dd']['lump']    = 0
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["tol"] = 1e-9
        # self.modeling_options['TowerSE']['frame3dd']['shift']   = 0.0
        # self.modeling_options['TowerSE']['frame3dd']['add_gravity'] = True

    def testPreFrame(self):

        # Test Land
        self.inputs["z_param"] = 10.0 * np.array([0.0, 3.0, 6.0])
        self.inputs["z_full"] = 10.0 * np.arange(0, 7)
        self.inputs["d_full"] = 6.0 * np.ones(self.inputs["z_full"].shape)
        self.inputs["mass"] = 1e5
        self.inputs["mI"] = np.r_[1e5, 1e5, 2e5, np.zeros(3)]
        self.inputs["mrho"] = np.array([-3.0, 0.0, 1.0])
        self.inputs["transition_piece_mass"] = 0.0
        self.inputs["transition_piece_cost"] = 0.0
        self.inputs["transition_piece_height"] = 0.0
        self.inputs["transition_piece_I"] = np.zeros(6)
        self.inputs["gravity_foundation_I"] = np.zeros(6)
        self.inputs["gravity_foundation_mass"] = 0.0
        self.inputs["suctionpile_depth"] = 0.0
        self.inputs["rna_F"] = 1e5 * np.arange(2,5)
        self.inputs["rna_M"] = 1e6 * np.arange(2,5)
        self.inputs["E"] = 1e9 * np.ones(2)
        self.inputs["G"] = 1e8 * np.ones(2)
        self.inputs["sigma_y"] = 1e8 * np.ones(2)

        myobj = tow.TowerPreFrame(n_height=3, n_refine=3, monopile=False)
        myobj.compute(self.inputs, self.outputs)

        npt.assert_equal(self.outputs["kidx"], np.array([0]))
        npt.assert_equal(self.outputs["kx"], np.array([RIGID]))
        npt.assert_equal(self.outputs["ky"], np.array([RIGID]))
        npt.assert_equal(self.outputs["kz"], np.array([RIGID]))
        npt.assert_equal(self.outputs["ktx"], np.array([RIGID]))
        npt.assert_equal(self.outputs["kty"], np.array([RIGID]))
        npt.assert_equal(self.outputs["ktz"], np.array([RIGID]))

        npt.assert_equal(self.outputs["midx"], np.zeros(2))
        npt.assert_equal(self.outputs["m"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhox"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhoy"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhoz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxx"], np.zeros(2))
        npt.assert_equal(self.outputs["mIyy"], np.zeros(2))
        npt.assert_equal(self.outputs["mIzz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxy"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIyz"], np.zeros(2))

        npt.assert_equal(self.outputs["plidx"], np.array([6]))
        npt.assert_equal(self.outputs["Fx"], np.array([2e5]))
        npt.assert_equal(self.outputs["Fy"], np.array([3e5]))
        npt.assert_equal(self.outputs["Fz"], np.array([4e5]))
        npt.assert_equal(self.outputs["Mxx"], np.array([2e6]))
        npt.assert_equal(self.outputs["Myy"], np.array([3e6]))
        npt.assert_equal(self.outputs["Mzz"], np.array([4e6]))

        # Test Monopile no springs, no GBF
        self.inputs["z_full"] = 10.0 * np.arange(-6, 7)
        self.inputs["d_full"] = 6.0 * np.ones(self.inputs["z_full"].shape)
        self.inputs["transition_piece_mass"] = 1e3
        self.inputs["transition_piece_cost"] = 1e4
        self.inputs["transition_piece_I"] = 1e3 * 9 * np.r_[0.5, 0.5, 1.0, np.zeros(3)]
        self.inputs["transition_piece_height"] = 10.0
        self.inputs["gravity_foundation_mass"] = 0.0  # 1e4
        self.inputs["suctionpile_depth"] = 30.0
        self.inputs["rna_F"] = 1e5 * np.arange(2,5)
        self.inputs["rna_M"] = 1e6 * np.arange(2,5)
        self.inputs["k_soil"] = (20.0 + np.arange(6))[np.newaxis, :] * np.ones((2, 6))
        self.inputs["z_soil"] = np.r_[-30.0, 0.0]

        myobj = tow.TowerPreFrame(n_height=5, n_refine=3, monopile=True, soil_springs=False)
        myobj.compute(self.inputs, self.outputs)

        npt.assert_equal(self.outputs["kidx"], np.arange(4))
        npt.assert_equal(self.outputs["kx"], RIGID)
        npt.assert_equal(self.outputs["ky"], RIGID)
        npt.assert_equal(self.outputs["kz"], RIGID)
        npt.assert_equal(self.outputs["ktx"], RIGID)
        npt.assert_equal(self.outputs["kty"], RIGID)
        npt.assert_equal(self.outputs["ktz"], RIGID)

        npt.assert_equal(self.outputs["midx"], np.array([7, 0]))
        npt.assert_equal(self.outputs["m"], np.array([1e3, 0.0]))
        npt.assert_equal(self.outputs["mrhox"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhoy"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhoz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxx"], np.array([1e3 * 9 * 0.5, 0]))
        npt.assert_equal(self.outputs["mIyy"], np.array([1e3 * 9 * 0.5, 0]))
        npt.assert_equal(self.outputs["mIzz"], np.array([1e3 * 9, 0]))
        npt.assert_equal(self.outputs["mIxy"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIyz"], np.zeros(2))

        npt.assert_equal(self.outputs["plidx"], np.array([12]))
        npt.assert_equal(self.outputs["Fx"], np.array([2e5]))
        npt.assert_equal(self.outputs["Fy"], np.array([3e5]))
        npt.assert_equal(self.outputs["Fz"], np.array([4e5]))
        npt.assert_equal(self.outputs["Mxx"], np.array([2e6]))
        npt.assert_equal(self.outputs["Myy"], np.array([3e6]))
        npt.assert_equal(self.outputs["Mzz"], np.array([4e6]))

        # Test Monopile springs, no GBF
        self.inputs["z_full"] = 10.0 * np.arange(-6, 7)
        self.inputs["d_full"] = 6.0 * np.ones(self.inputs["z_full"].shape)
        self.inputs["transition_piece_mass"] = 1e3
        self.inputs["transition_piece_cost"] = 1e4
        self.inputs["transition_piece_I"] = 1e3 * 9 * np.r_[0.5, 0.5, 1.0, np.zeros(3)]
        self.inputs["transition_piece_height"] = 10.0
        self.inputs["gravity_foundation_mass"] = 0.0  # 1e4
        self.inputs["suctionpile_depth"] = 30.0
        self.inputs["rna_F"] = 1e5 * np.arange(2,5)
        self.inputs["rna_M"] = 1e6 * np.arange(2,5)
        self.inputs["k_soil"] = (20.0 + np.arange(6))[np.newaxis, :] * np.ones((2, 6))
        self.inputs["z_soil"] = np.r_[-30.0, 0.0]

        myobj = tow.TowerPreFrame(n_height=5, n_refine=3, monopile=True, soil_springs=True)
        myobj.compute(self.inputs, self.outputs)

        npt.assert_equal(self.outputs["kidx"], np.arange(4))
        npt.assert_equal(self.outputs["kx"], 20.0)
        npt.assert_equal(self.outputs["ky"], 22.0)
        npt.assert_equal(self.outputs["kz"], np.r_[24.0, np.zeros(3)])
        npt.assert_equal(self.outputs["ktx"], 21.0)
        npt.assert_equal(self.outputs["kty"], 23.0)
        npt.assert_equal(self.outputs["ktz"], 25.0)

        npt.assert_equal(self.outputs["midx"], np.array([7, 0]))
        npt.assert_equal(self.outputs["m"], np.array([1e3, 0.0]))
        npt.assert_equal(self.outputs["mrhox"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhoy"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhoz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxx"], np.array([1e3 * 9 * 0.5, 0]))
        npt.assert_equal(self.outputs["mIyy"], np.array([1e3 * 9 * 0.5, 0]))
        npt.assert_equal(self.outputs["mIzz"], np.array([1e3 * 9, 0]))
        npt.assert_equal(self.outputs["mIxy"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIyz"], np.zeros(2))

        npt.assert_equal(self.outputs["plidx"], np.array([12]))
        npt.assert_equal(self.outputs["Fx"], np.array([2e5]))
        npt.assert_equal(self.outputs["Fy"], np.array([3e5]))
        npt.assert_equal(self.outputs["Fz"], np.array([4e5]))
        npt.assert_equal(self.outputs["Mxx"], np.array([2e6]))
        npt.assert_equal(self.outputs["Myy"], np.array([3e6]))
        npt.assert_equal(self.outputs["Mzz"], np.array([4e6]))

        # Test Monopile with GBF- TODO: THESE REACTIONS NEED THOUGHT
        self.inputs["z_full"] = 10.0 * np.arange(-6, 7)
        self.inputs["d_full"] = 6.0 * np.ones(self.inputs["z_full"].shape)
        self.inputs["transition_piece_mass"] = 1e3
        self.inputs["transition_piece_cost"] = 1e4
        self.inputs["transition_piece_height"] = 10.0
        self.inputs["transition_piece_I"] = 1e3 * 9 * np.r_[0.5, 0.5, 1.0, np.zeros(3)]
        self.inputs["gravity_foundation_I"] = 0.5 * 1e4 * 9 * np.r_[0.5, 0.5, 1.0, np.zeros(3)]
        self.inputs["gravity_foundation_mass"] = 1e4
        self.inputs["suctionpile_depth"] = 0.0
        self.inputs["rna_F"] = 1e5 * np.arange(2,5)
        self.inputs["rna_M"] = 1e6 * np.arange(2,5)
        self.inputs["k_soil"] = (20.0 + np.arange(6))[np.newaxis, :] * np.ones((2, 6))
        self.inputs["z_soil"] = np.r_[-30.0, 0.0]

        myobj = tow.TowerPreFrame(n_height=5, n_refine=3, monopile=True, gravity_foundation=True)
        myobj.compute(self.inputs, self.outputs)

        npt.assert_equal(self.outputs["kidx"], np.array([0]))
        npt.assert_equal(self.outputs["kx"], np.array([RIGID]))
        npt.assert_equal(self.outputs["ky"], np.array([RIGID]))
        npt.assert_equal(self.outputs["kz"], np.array([RIGID]))
        npt.assert_equal(self.outputs["ktx"], np.array([RIGID]))
        npt.assert_equal(self.outputs["kty"], np.array([RIGID]))
        npt.assert_equal(self.outputs["ktz"], np.array([RIGID]))

        npt.assert_equal(self.outputs["midx"], np.array([7, 0]))
        npt.assert_equal(self.outputs["m"], np.array([1e3, 1e4]))
        npt.assert_equal(self.outputs["mrhox"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhoy"], np.zeros(2))
        npt.assert_equal(self.outputs["mrhoz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxx"], np.array([1e3 * 9 * 0.5, 1e4 * 9 * 0.25]))
        npt.assert_equal(self.outputs["mIyy"], np.array([1e3 * 9 * 0.5, 1e4 * 9 * 0.25]))
        npt.assert_equal(self.outputs["mIzz"], np.array([1e3 * 9, 1e4 * 9 * 0.5]))
        npt.assert_equal(self.outputs["mIxy"], np.zeros(2))
        npt.assert_equal(self.outputs["mIxz"], np.zeros(2))
        npt.assert_equal(self.outputs["mIyz"], np.zeros(2))

        npt.assert_equal(self.outputs["plidx"], np.array([12]))
        npt.assert_equal(self.outputs["Fx"], np.array([2e5]))
        npt.assert_equal(self.outputs["Fy"], np.array([3e5]))
        npt.assert_equal(self.outputs["Fz"], np.array([4e5]))
        npt.assert_equal(self.outputs["Mxx"], np.array([2e6]))
        npt.assert_equal(self.outputs["Myy"], np.array([3e6]))
        npt.assert_equal(self.outputs["Mzz"], np.array([4e6]))


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestStruct))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import copy

import numpy as np
import openmdao.api as om
import wisdem.commonse.frustum as frustum
import wisdem.commonse.utilities as util
import wisdem.commonse.manufacturing as manufacture
from wisdem.towerse import NREFINE, eps, get_nfull


class DiscretizationYAML(om.ExplicitComponent):
    """
    Process some of the tower YAML inputs.

    Parameters
    ----------
    tower_s : numpy array[n_height_tow]
        1D array of the non-dimensional grid defined along the tower axis (0-tower base,
        1-tower top)
    tower_layer_materials : list of strings
        1D array of the names of the materials of each layer modeled in the tower
        structure.
    tower_layer_thickness : numpy array[n_layers_tow, n_height_tow], [m]
        2D array of the thickness of the layers of the tower structure. The first
        dimension represents each layer, the second dimension represents each piecewise-
        constant entry of the tower sections.
    tower_height : float, [m]
        Scalar of the tower height computed along the z axis.
    tower_outer_diameter_in : numpy array[n_height_tow], [m]
        cylinder diameter at corresponding locations
    tower_outfitting_factor : float
        Multiplier that accounts for secondary structure mass inside of cylinder
    monopile_s : numpy array[n_height_mon]
        1D array of the non-dimensional grid defined along the tower axis (0-tower base,
        1-tower top)
    monopile_layer_materials : list of strings
        1D array of the names of the materials of each layer modeled in the tower
        structure.
    monopile_layer_thickness : numpy array[n_layers_mon, n_height_mon], [m]
        2D array of the thickness of the layers of the tower structure. The first
        dimension represents each layer, the second dimension represents each piecewise-
        constant entry of the tower sections.
    monopile_height : float, [m]
        Scalar of the tower height computed along the z axis.
    monopile_outer_diameter_in : numpy array[n_height_tow], [m]
        cylinder diameter at corresponding locations
    monopile_outfitting_factor : float
        Multiplier that accounts for secondary structure mass inside of cylinder
    material_names : list of strings
        1D array of names of materials.
    E_mat : numpy array[n_mat, 3], [Pa]
        2D array of the Youngs moduli of the materials. Each row represents a material,
        the three columns represent E11, E22 and E33.
    G_mat : numpy array[n_mat, 3], [Pa]
        2D array of the shear moduli of the materials. Each row represents a material,
        the three columns represent G12, G13 and G23.
    sigma_y_mat : numpy array[n_mat], [Pa]
        2D array of the yield strength of the materials. Each row represents a material,
        the three columns represent Xt12, Xt13 and Xt23.
    rho_mat : numpy array[n_mat], [kg/m**3]
        1D array of the density of the materials. For composites, this is the density of
        the laminate.
    unit_cost_mat : numpy array[n_mat], [USD/kg]
        1D array of the unit costs of the materials.

    Returns
    -------
    tower_section_height : numpy array[n_height-1], [m]
        parameterized section heights along cylinder
    tower_outer_diameter : numpy array[n_height], [m]
        cylinder diameter at corresponding locations
    tower_wall_thickness : numpy array[n_height-1], [m]
        shell thickness at corresponding locations
    suctionpile_depth : float, [m]
        Depth of monopile below sea floor
    outfitting_factor : numpy array[n_height-1]
        Multiplier that accounts for secondary structure mass inside of cylinder
    E : numpy array[n_height-1], [Pa]
        Isotropic Youngs modulus of the materials along the tower sections.
    G : numpy array[n_height-1], [Pa]
        Isotropic shear modulus of the materials along the tower sections.
    sigma_y : numpy array[n_height-1], [Pa]
        Isotropic yield strength of the materials along the tower sections.
    rho : numpy array[n_height-1], [kg/m**3]
        Density of the materials along the tower sections.
    unit_cost : numpy array[n_height-1], [USD/kg]
        Unit costs of the materials along the tower sections.
    outfitting_factor : numpy array[n_height-1]
        Additional outfitting multiplier in each section

    """

    def initialize(self):
        self.options.declare("n_height_tower")
        self.options.declare("n_height_monopile")
        self.options.declare("n_layers_tower")
        self.options.declare("n_layers_monopile")
        self.options.declare("n_mat")

    def setup(self):
        n_height_tow = self.options["n_height_tower"]
        n_height_mon = self.options["n_height_monopile"]
        n_layers_tow = self.options["n_layers_tower"]
        n_layers_mon = self.options["n_layers_monopile"]
        n_mat = self.options["n_mat"]
        if n_height_mon > 0:
            n_height = n_height_tow + n_height_mon - 1  # Should have one overlapping point
        else:
            n_height = n_height_tow

        # Inputs here are the outputs from the Tower component in load_IEA_yaml
        # TODO: Use reference axis and curvature, s, instead of assuming everything is vertical on z
        self.add_input("tower_s", val=np.zeros(n_height_tow))
        self.add_discrete_input("tower_layer_materials", val=n_layers_tow * [""])
        self.add_input("tower_layer_thickness", val=np.zeros((n_layers_tow, n_height_tow)), units="m")
        self.add_input("tower_height", val=0.0, units="m")
        self.add_input("tower_foundation_height", val=0.0, units="m")
        self.add_input("tower_outer_diameter_in", np.zeros(n_height_tow), units="m")
        self.add_input("tower_outfitting_factor", val=0.0)
        self.add_input("monopile_s", val=np.zeros(n_height_mon))
        self.add_discrete_input("monopile_layer_materials", val=n_layers_tow * [""])
        self.add_input("monopile_layer_thickness", val=np.zeros((n_layers_mon, n_height_mon)), units="m")
        self.add_input("monopile_foundation_height", val=0.0, units="m")
        self.add_input("monopile_height", val=0.0, units="m")
        self.add_input("monopile_outer_diameter_in", np.zeros(n_height_mon), units="m")
        self.add_input("monopile_outfitting_factor", val=0.0)
        self.add_discrete_input("material_names", val=n_mat * [""])
        self.add_input("water_depth", val=0.0, units="m")
        self.add_input("E_mat", val=np.zeros([n_mat, 3]), units="Pa")
        self.add_input("G_mat", val=np.zeros([n_mat, 3]), units="Pa")
        self.add_input("sigma_y_mat", val=np.zeros(n_mat), units="Pa")
        self.add_input("sigma_ult_mat", val=np.zeros([n_mat, 3]), units="Pa")
        self.add_input("wohler_exp_mat", val=np.zeros(n_mat))
        self.add_input("wohler_A_mat", val=np.zeros(n_mat))
        self.add_input("rho_mat", val=np.zeros(n_mat), units="kg/m**3")
        self.add_input("unit_cost_mat", val=np.zeros(n_mat), units="USD/kg")

        self.add_output("tower_section_height", val=np.zeros(n_height - 1), units="m")
        self.add_output("tower_outer_diameter", val=np.zeros(n_height), units="m")
        self.add_output("tower_wall_thickness", val=np.zeros(n_height - 1), units="m")
        self.add_output("transition_piece_height", 0.0, units="m")
        self.add_output("suctionpile_depth", 0.0, units="m")
        self.add_output("outfitting_factor", val=np.zeros(n_height - 1))
        self.add_output("E", val=np.zeros(n_height - 1), units="Pa")
        self.add_output("G", val=np.zeros(n_height - 1), units="Pa")
        self.add_output("sigma_y", val=np.zeros(n_height - 1), units="Pa")
        self.add_output("sigma_ult", val=np.zeros(n_height - 1), units="Pa")
        self.add_output("wohler_exp", val=np.zeros(n_height - 1))
        self.add_output("wohler_A", val=np.zeros(n_height - 1))
        self.add_output("rho", val=np.zeros(n_height - 1), units="kg/m**3")
        self.add_output("unit_cost", val=np.zeros(n_height - 1), units="USD/kg")
        self.add_output("z_start", 0.0, units="m")

        # self.declare_partials('*', '*', method='fd')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack dimensions
        n_height_tow = self.options["n_height_tower"]
        n_height_mon = self.options["n_height_monopile"]
        n_layers_tow = self.options["n_layers_tower"]
        n_layers_mon = self.options["n_layers_monopile"]
        n_height = n_height_tow if n_height_mon == 0 else n_height_tow + n_height_mon - 1

        # Unpack values
        h_mon = inputs["monopile_height"]
        h_tow = inputs["tower_height"]
        s_mon = inputs["monopile_s"]
        s_tow = inputs["tower_s"]
        lthick_mon = inputs["monopile_layer_thickness"]
        lthick_tow = inputs["tower_layer_thickness"]
        lmat_mon = copy.copy(discrete_inputs["monopile_layer_materials"])
        lmat_tow = copy.copy(discrete_inputs["tower_layer_materials"])
        fh_tow = inputs["tower_foundation_height"]
        fh_mon = inputs["monopile_foundation_height"]
        water_depth = inputs["water_depth"]

        # Enforce constant tower thickness per section, assuming rolling of a flat steel plate
        # Might not have to assume this with concrete, but can account for it in input values if necessary
        lthick_tow = 0.5 * (lthick_tow[:, :-1] + lthick_tow[:, 1:])
        if n_height_mon > 0:
            lthick_mon = 0.5 * (lthick_mon[:, :-1] + lthick_mon[:, 1:])

        outputs["transition_piece_height"] = fh_tow

        if n_height_mon > 0:
            if np.abs(fh_tow - fh_mon - h_mon) > 1.0:
                print(
                    "WARNING: Monopile length is not consistent with transition piece height and monopile base height\n",
                    "         Determining new base height value . . .",
                )
            outputs["z_start"] = fh_tow - h_mon

            pile = h_mon - fh_tow - water_depth
            outputs["suctionpile_depth"] = pile
            # Ensure that we have only one segment for pile, a current limitation
            if pile > 0:
                s1 = pile / h_mon
                icheck = np.where(s_mon > s1 + 1e-3)[0][0]
                s_mon = np.r_[0.0, np.linspace(s1, s_mon[icheck], icheck).flatten(), s_mon[(icheck + 1) :].flatten()]

            # Last monopile point and first tower point are the same
            outputs["tower_section_height"] = np.r_[np.diff(h_mon * s_mon), np.diff(h_tow * s_tow)]
            outputs["outfitting_factor"] = np.r_[
                inputs["monopile_outfitting_factor"] * np.ones(n_height_mon - 1),
                inputs["tower_outfitting_factor"] * np.ones(n_height_tow - 1),
            ]
            outputs["tower_outer_diameter"] = np.r_[
                inputs["monopile_outer_diameter_in"], inputs["tower_outer_diameter_in"][1:]
            ]

            # Combine layers into one structure
            layer_mat = []
            twall = np.zeros((1, n_height - 1))
            for k in range(n_layers_mon):
                ilayer = np.zeros(n_height - 1)

                ilayer[: (n_height_mon - 1)] = lthick_mon[k, :]

                imat_mon = lmat_mon[k]
                layer_mat.append(imat_mon)

                if imat_mon in lmat_tow:
                    ktow = lmat_tow.index(imat_mon)
                    ilayer[(n_height_mon - 1) :] = lthick_tow[ktow, :]

                    # Remove from listing so we don't double count later
                    lmat_tow.pop(ktow)
                    lthick_tow = np.delete(lthick_tow, [ktow], axis=0)

                twall = np.vstack((twall, ilayer))

            # If there any uncounted tower layers, add them in
            n_layers_tow = len(lmat_tow)
            for k in range(n_layers_tow):
                ilayer = np.zeros(n_height - 1)
                ilayer[(n_height_mon - 1) :] = lthick_tow[k, :]
                twall = np.vstack((twall, ilayer))
                imat = lmat_tow[k]
                layer_mat.append(imat)

            twall = np.delete(twall, [0], axis=0)
            outputs["tower_wall_thickness"] = np.sum(twall, axis=0)

        else:
            outputs["tower_section_height"] = np.diff(h_tow * s_tow)
            outputs["tower_wall_thickness"] = np.sum(lthick_tow, axis=0)
            outputs["outfitting_factor"] = inputs["tower_outfitting_factor"] * np.ones(n_height - 1)
            outputs["tower_outer_diameter"] = inputs["tower_outer_diameter_in"]
            twall = lthick_tow
            layer_mat = discrete_inputs["tower_layer_materials"]
            outputs["z_start"] = fh_tow
            outputs["suctionpile_depth"] = 0.0

        # Check to make sure we have good values
        if np.any(outputs["tower_section_height"] <= 0.0):
            raise ValueError("Section height values must be greater than zero, " + str(outputs["tower_section_height"]))
        if np.any(outputs["tower_wall_thickness"] <= 0.0):
            raise ValueError("Wall thickness values must be greater than zero, " + str(outputs["tower_wall_thickness"]))
        if np.any(outputs["tower_outer_diameter"] <= 0.0):
            raise ValueError("Diameter values must be greater than zero, " + str(outputs["tower_outer_diameter"]))

        # DETERMINE MATERIAL PROPERTIES IN EACH SECTION
        # Convert to isotropic material
        E = np.mean(inputs["E_mat"], axis=1)
        G = np.mean(inputs["G_mat"], axis=1)
        sigy = inputs["sigma_y_mat"]
        sigu = inputs["sigma_ult_mat"].min(axis=1)
        m = inputs["wohler_exp_mat"]
        A = inputs["wohler_A_mat"]
        rho = inputs["rho_mat"]
        cost = inputs["unit_cost_mat"]
        mat_names = discrete_inputs["material_names"]

        # Initialize sectional data
        E_param = np.zeros(twall.shape)
        G_param = np.zeros(twall.shape)
        sigy_param = np.zeros(twall.shape)
        sigu_param = np.zeros(twall.shape)
        m_param = np.zeros(twall.shape)
        A_param = np.zeros(twall.shape)
        rho_param = np.zeros(n_height - 1)
        cost_param = np.zeros(n_height - 1)

        # Loop over materials and associate it with its thickness
        for k in range(len(layer_mat)):
            # Get the material name for this layer
            iname = layer_mat[k]

            # Get the index into the material list
            imat = mat_names.index(iname)

            imass = rho[imat] * twall[k, :]

            # For density, take mass weighted layer
            rho_param += imass

            # For cost, take mass weighted layer
            cost_param += imass * cost[imat]

            # Store the value associated with this thickness
            E_param[k, :] = E[imat]
            G_param[k, :] = G[imat]
            sigy_param[k, :] = sigy[imat]
            sigu_param[k, :] = sigu[imat]
            m_param[k, :] = m[imat]
            A_param[k, :] = A[imat]

        # Mass weighted cost (should really weight by radius too)
        cost_param /= rho_param

        # Thickness weighted density (should really weight by radius too)
        rho_param /= twall.sum(axis=0)

        # Mixtures of material properties: https://en.wikipedia.org/wiki/Rule_of_mixtures

        # Volume fraction
        vol_frac = twall / twall.sum(axis=0)[np.newaxis, :]

        # Average of upper and lower bounds
        E_param = 0.5 * np.sum(vol_frac * E_param, axis=0) + 0.5 / np.sum(vol_frac / E_param, axis=0)
        G_param = 0.5 * np.sum(vol_frac * G_param, axis=0) + 0.5 / np.sum(vol_frac / G_param, axis=0)
        sigy_param = 0.5 * np.sum(vol_frac * sigy_param, axis=0) + 0.5 / np.sum(vol_frac / sigy_param, axis=0)
        sigu_param = 0.5 * np.sum(vol_frac * sigu_param, axis=0) + 0.5 / np.sum(vol_frac / sigu_param, axis=0)
        m_param = 0.5 * np.sum(vol_frac * m_param, axis=0) + 0.5 / np.sum(vol_frac / m_param, axis=0)
        A_param = 0.5 * np.sum(vol_frac * A_param, axis=0) + 0.5 / np.sum(vol_frac / A_param, axis=0)

        # Store values
        outputs["E"] = E_param
        outputs["G"] = G_param
        outputs["rho"] = rho_param
        outputs["sigma_y"] = sigy_param
        outputs["sigma_ult"] = sigu_param
        outputs["wohler_exp"] = m_param
        outputs["wohler_A"] = A_param
        outputs["unit_cost"] = cost_param


class CylinderDiscretization(om.ExplicitComponent):
    """
    Discretize geometry into finite element nodes

    Parameters
    ----------
    foundation_height : float, [m]
        starting height of tower
    section_height : numpy array[nPoints-1], [m]
        parameterized section heights along cylinder
    diameter : numpy array[nPoints], [m]
        cylinder diameter at corresponding locations
    wall_thickness : numpy array[nPoints-1], [m]
        shell thickness at corresponding locations

    Returns
    -------
    z_param : numpy array[nPoints], [m]
        parameterized locations along cylinder, linear lofting between
    z_full : numpy array[nFull], [m]
        locations along cylinder
    d_full : numpy array[nFull], [m]
        cylinder diameter at corresponding locations
    t_full : numpy array[nFull-1], [m]
        shell thickness at corresponding locations

    """

    """discretize geometry into finite element nodes"""

    def initialize(self):
        self.options.declare("nPoints")
        self.options.declare("nRefine", default=NREFINE)
        self.options.declare("nPin", default=0)

    def setup(self):
        nPoints = self.options["nPoints"]
        nRefine = np.round(self.options["nRefine"])
        nFull = int(nRefine * (nPoints - 1) + 1)

        self.add_input("foundation_height", val=0.0, units="m")
        self.add_input("section_height", np.zeros(nPoints - 1), units="m")
        self.add_input("diameter", np.zeros(nPoints), units="m")
        self.add_input("wall_thickness", np.zeros(nPoints - 1), units="m")

        self.add_output("z_param", np.zeros(nPoints), units="m")
        self.add_output("z_full", np.zeros(nFull), units="m")
        self.add_output("d_full", np.zeros(nFull), units="m")
        self.add_output("t_full", np.zeros(nFull - 1), units="m")

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):
        # Check to make sure we have good values
        if np.any(inputs["section_height"] <= 0.0):
            raise ValueError("Section height values must be greater than zero, " + str(inputs["section_height"]))
        if np.any(inputs["wall_thickness"] <= 0.0):
            raise ValueError("Wall thickness values must be greater than zero, " + str(inputs["wall_thickness"]))
        if np.any(inputs["diameter"] <= 0.0):
            raise ValueError("Diameter values must be greater than zero, " + str(inputs["diameter"]))

        nRefine = int(np.round(self.options["nRefine"]))
        z_param = float(inputs["foundation_height"]) + np.r_[0.0, np.cumsum(inputs["section_height"].flatten())]

        # Have to regine each element one at a time so that we preserve input nodes
        z_full = np.array([])
        for k in range(z_param.size - 1):
            zref = np.linspace(z_param[k], z_param[k + 1], nRefine + 1)
            z_full = np.append(z_full, zref)
        z_full = np.unique(z_full)

        outputs["z_full"] = z_full
        outputs["d_full"] = np.interp(z_full, z_param, inputs["diameter"])
        z_section = 0.5 * (z_full[:-1] + z_full[1:])
        outputs["t_full"] = util.sectionalInterp(z_section, z_param, inputs["wall_thickness"])
        outputs["z_param"] = z_param


class TowerDiscretization(om.ExplicitComponent):
    """
    Compute the full arrays for some measures along the tower by interpolating.

    Parameters
    ----------
    hub_height : float, [m]
        diameter at tower base
    z_param : numpy array[n_height], [m]
        parameterized locations along tower, linear lofting between
    z_full : numpy array[nFull], [m]
        parameterized locations along tower, linear lofting between
    tower_outer_diameter : numpy array[n_height], [m]
        cylinder diameter at corresponding locations
    rho : numpy array[n_height-1], [kg/m**3]
        Density of the materials along the tower sections.
    unit_cost : numpy array[n_height-1], [USD/kg]
        Unit costs of the materials along the tower sections.
    outfitting_factor : numpy array[n_height-1]
        Multiplier that accounts for secondary structure mass inside of cylinder
    E : numpy array[n_height-1], [Pa]
        Isotropic Youngs modulus of the materials along the tower sections.
    G : numpy array[n_height-1], [Pa]
        Isotropic shear modulus of the materials along the tower sections.
    sigma_y : numpy array[n_height-1], [Pa]
        Isotropic yield strength of the materials along the tower sections.
    Az : numpy array[nFull-1], [m**2]
        cross-sectional area
    Jz : numpy array[nFull-1], [m**4]
        polar moment of inertia
    Ixx : numpy array[nFull-1], [m**4]
        area moment of inertia about x-axis
    Iyy : numpy array[nFull-1], [m**4]
        area moment of inertia about y-axis

    Returns
    -------
    height_constraint : float, [m]
        mismatch between tower height and desired hub_height
    rho_full : numpy array[nFull-1], [kg/m**3]
        Density of the materials along the tower sections.
    unit_cost_full : numpy array[nFull-1], [USD/kg]
        Unit costs of the materials along the tower sections.
    outfitting_full : numpy array[nFull-1]
        Multiplier that accounts for secondary structure mass inside of cylinder
    E_full : numpy array[nFull-1], [Pa]
        Isotropic Youngs modulus of the materials along the tower sections.
    G_full : numpy array[nFull-1], [Pa]
        Isotropic shear modulus of the materials along the tower sections.
    sigma_y_full : numpy array[nFull-1], [Pa]
        Isotropic yield strength of the materials along the tower sections.
    axial_load2stress : numpy array[nFull-1,6], [m**2]
        Linear conversion factors between loads [Fx-z; Mx-z] and axial stress
    shear_load2stress : numpy array[nFull-1,6], [m**2]
        Linear conversion factors between loads [Fx-z; Mx-z] and shear stress

    """

    def initialize(self):
        self.options.declare("n_height")
        self.options.declare("n_refine")
        self.options.declare("modeling_options")

    def setup(self):
        n_height = self.options["n_height"]
        n_refine = self.options["n_refine"]
        nFull = get_nfull(n_height, nref=n_refine)

        self.add_input("hub_height", val=0.0, units="m")
        self.add_input("z_param", np.zeros(n_height), units="m")
        self.add_input("z_full", val=np.zeros(nFull), units="m")
        self.add_input("tower_outer_diameter", val=np.zeros(n_height), units="m")
        self.add_input("rho", val=np.zeros(n_height - 1), units="kg/m**3")
        self.add_input("unit_cost", val=np.zeros(n_height - 1), units="USD/kg")
        self.add_input("outfitting_factor", val=np.zeros(n_height - 1))
        self.add_input("E", val=np.zeros(n_height - 1), units="Pa")
        self.add_input("G", val=np.zeros(n_height - 1), units="Pa")
        self.add_input("sigma_y", val=np.zeros(n_height - 1), units="Pa")

        self.add_input("Az", np.zeros(nFull - 1), units="m**2")
        self.add_input("Asx", np.zeros(nFull - 1), units="m**2")
        self.add_input("Asy", np.zeros(nFull - 1), units="m**2")
        self.add_input("Jz", np.zeros(nFull - 1), units="m**4")
        self.add_input("Ixx", np.zeros(nFull - 1), units="m**4")
        self.add_input("Iyy", np.zeros(nFull - 1), units="m**4")

        self.add_output("height_constraint", val=0.0, units="m")
        self.add_output("rho_full", val=np.zeros(nFull - 1), units="kg/m**3")
        self.add_output("unit_cost_full", val=np.zeros(nFull - 1), units="USD/kg")
        self.add_output("outfitting_full", val=np.zeros(nFull - 1))
        self.add_output("E_full", val=np.zeros(nFull - 1), units="Pa")
        self.add_output("G_full", val=np.zeros(nFull - 1), units="Pa")
        self.add_output("sigma_y_full", val=np.zeros(nFull - 1), units="Pa")

        # Tower Distributed Beam Properties (properties needed for ElastoDyn (OpenFAST) inputs or BModes inputs for verification purposes)
        self.add_output("sec_loc", np.zeros(n_height - 1), desc="normalized sectional location")
        self.add_output("str_tw", np.zeros(n_height - 1), units="deg", desc="structural twist of section")
        self.add_output("tw_iner", np.zeros(n_height - 1), units="deg", desc="inertial twist of section")
        self.add_output("mass_den", np.zeros(n_height - 1), units="kg/m", desc="sectional mass per unit length")
        self.add_output(
            "foreaft_iner",
            np.zeros(n_height - 1),
            units="kg*m",
            desc="sectional fore-aft intertia per unit length about the Y_G inertia axis",
        )
        self.add_output(
            "sideside_iner",
            np.zeros(n_height - 1),
            units="kg*m",
            desc="sectional side-side intertia per unit length about the Y_G inertia axis",
        )
        self.add_output(
            "foreaft_stff",
            np.zeros(n_height - 1),
            units="N*m**2",
            desc="sectional fore-aft bending stiffness per unit length about the Y_E elastic axis",
        )
        self.add_output(
            "sideside_stff",
            np.zeros(n_height - 1),
            units="N*m**2",
            desc="sectional side-side bending stiffness per unit length about the Y_E elastic axis",
        )
        self.add_output("tor_stff", np.zeros(n_height - 1), units="N*m**2", desc="sectional torsional stiffness")
        self.add_output("axial_stff", np.zeros(n_height - 1), units="N", desc="sectional axial stiffness")
        self.add_output("cg_offst", np.zeros(n_height - 1), units="m", desc="offset from the sectional center of mass")
        self.add_output("sc_offst", np.zeros(n_height - 1), units="m", desc="offset from the sectional shear center")
        self.add_output("tc_offst", np.zeros(n_height - 1), units="m", desc="offset from the sectional tension center")
        self.add_output("axial_load2stress", val=np.zeros((n_height - 1, 6)), units="m**2")
        self.add_output("shear_load2stress", val=np.zeros((n_height - 1, 6)), units="m**2")

        self.declare_partials("height_constraint", ["hub_height", "z_param"], method="fd")
        self.declare_partials("outfitting_full", ["outfitting_factor"], method="fd")
        self.declare_partials("rho_full", ["rho"], method="fd")
        self.declare_partials("unit_cost_full", ["unit_cost"], method="fd")

    def compute(self, inputs, outputs):
        z_full = inputs["z_full"]
        z_param = inputs["z_param"]
        z_section = 0.5 * (z_full[:-1] + z_full[1:])

        outputs["height_constraint"] = inputs["hub_height"] - z_param[-1]
        outputs["rho_full"] = util.sectionalInterp(z_section, z_param, inputs["rho"])
        outputs["outfitting_full"] = util.sectionalInterp(z_section, z_param, inputs["outfitting_factor"])
        outputs["unit_cost_full"] = util.sectionalInterp(z_section, z_param, inputs["unit_cost"])
        outputs["E_full"] = util.sectionalInterp(z_section, z_param, inputs["E"])
        outputs["G_full"] = util.sectionalInterp(z_section, z_param, inputs["G"])
        outputs["sigma_y_full"] = util.sectionalInterp(z_section, z_param, inputs["sigma_y"])

        # Unpack for Elastodyn
        z = 0.5 * (z_param[:-1] + z_param[1:])
        rho = inputs["rho"]
        E = inputs["E"]
        G = inputs["G"]
        of = inputs["outfitting_factor"]
        Az = util.sectionalInterp(z, z_full, inputs["Az"])
        Asx = util.sectionalInterp(z, z_full, inputs["Asx"])
        Asy = util.sectionalInterp(z, z_full, inputs["Asy"])
        Ixx = util.sectionalInterp(z, z_full, inputs["Ixx"])
        Iyy = util.sectionalInterp(z, z_full, inputs["Iyy"])
        Jz = util.sectionalInterp(z, z_full, inputs["Jz"])
        outputs["sec_loc"] = (z - z[0]) / (z[-1] - z[0])
        outputs["mass_den"] = rho * Az * of
        outputs["foreaft_iner"] = rho * Ixx
        outputs["sideside_iner"] = rho * Iyy
        outputs["foreaft_stff"] = E * Ixx
        outputs["sideside_stff"] = E * Iyy
        outputs["tor_stff"] = G * Jz
        outputs["axial_stff"] = E * Az

        d_sec, _ = util.nodal2sectional(inputs["tower_outer_diameter"])
        r_sec = 0.5 * d_sec
        ax_load2stress = np.zeros((d_sec.size, 6))
        ax_load2stress[:, 2] = 1.0 / Az
        ax_load2stress[:, 3] = r_sec / Ixx
        ax_load2stress[:, 4] = r_sec / Iyy
        sh_load2stress = np.zeros((d_sec.size, 6))
        sh_load2stress[:, 0] = 1.0 / Asx
        sh_load2stress[:, 1] = 1.0 / Asy
        sh_load2stress[:, 5] = r_sec / Jz
        outputs["axial_load2stress"] = ax_load2stress
        outputs["shear_load2stress"] = sh_load2stress


class CylinderMass(om.ExplicitComponent):
    """
    Compute cylinder cost and mass properties

    Parameters
    ----------
    d_full : numpy array[nPoints], [m]
        cylinder diameter at corresponding locations
    t_full : numpy array[nPoints-1], [m]
        shell thickness at corresponding locations
    z_full : numpy array[nPoints], [m]
        parameterized locations along cylinder, linear lofting between
    rho : numpy array[nPoints-1], [kg/m**3]
        material density
    outfitting_factor : numpy array[nPoints-1]
        Multiplier that accounts for secondary structure mass inside of cylinder
    material_cost_rate : numpy array[nPoints-1], [USD/kg]
        Raw material cost rate: steel $1.1/kg, aluminum $3.5/kg
    labor_cost_rate : float, [USD/min]
        Labor cost rate
    painting_cost_rate : float, [USD/m/m]
        Painting / surface finishing cost rate

    Returns
    -------
    cost : float, [USD]
        Total cylinder cost
    mass : numpy array[nPoints-1], [kg]
        Total cylinder mass
    center_of_mass : float, [m]
        z-position of center of mass of cylinder
    section_center_of_mass : numpy array[nPoints-1], [m]
        z position of center of mass of each can in the cylinder
    I_base : numpy array[6], [kg*m**2]
        mass moment of inertia of cylinder about base [xx yy zz xy xz yz]

    """

    def initialize(self):
        self.options.declare("nPoints")

    def setup(self):
        nPoints = self.options["nPoints"]

        self.add_input("d_full", val=np.zeros(nPoints), units="m")
        self.add_input("t_full", val=np.zeros(nPoints - 1), units="m")
        self.add_input("z_full", val=np.zeros(nPoints), units="m")
        self.add_input("rho", val=np.zeros(nPoints - 1), units="kg/m**3")
        self.add_input("outfitting_factor", val=np.zeros(nPoints - 1))
        self.add_input("material_cost_rate", val=np.zeros(nPoints - 1), units="USD/kg")
        self.add_input("labor_cost_rate", 0.0, units="USD/min")
        self.add_input("painting_cost_rate", 0.0, units="USD/m/m")

        self.add_output("cost", val=0.0, units="USD")
        self.add_output("mass", val=np.zeros(nPoints - 1), units="kg")
        self.add_output("center_of_mass", val=0.0, units="m")
        self.add_output("section_center_of_mass", val=np.zeros(nPoints - 1), units="m")
        self.add_output("I_base", np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs):
        # Unpack variables for thickness and average radius at each can interface
        twall = inputs["t_full"]
        Rb = 0.5 * inputs["d_full"][:-1]
        Rt = 0.5 * inputs["d_full"][1:]
        zz = inputs["z_full"]
        H = np.diff(zz)
        rho = inputs["rho"]
        coeff = inputs["outfitting_factor"]
        coeff = coeff + np.where(coeff < 1.0, 1.0, 0.0)

        # Total mass of cylinder
        V_shell = frustum.frustumShellVol(Rb, Rt, twall, H)
        mass = outputs["mass"] = coeff * rho * V_shell

        # Center of mass of each can/section
        cm_section = zz[:-1] + frustum.frustumShellCG(Rb, Rt, twall, H)
        outputs["section_center_of_mass"] = cm_section

        # Center of mass of cylinder
        V_shell += eps
        outputs["center_of_mass"] = np.dot(V_shell, cm_section) / V_shell.sum()

        # Moments of inertia
        Izz_section = coeff * rho * frustum.frustumShellIzz(Rb, Rt, twall, H)
        Ixx_section = Iyy_section = coeff * rho * frustum.frustumShellIxx(Rb, Rt, twall, H)

        # Sum up each cylinder section using parallel axis theorem
        I_base = np.zeros((3, 3))
        for k in range(Izz_section.size):
            R = np.array([0.0, 0.0, cm_section[k] - zz[0]])
            Icg = util.assembleI([Ixx_section[k], Iyy_section[k], Izz_section[k], 0.0, 0.0, 0.0])

            I_base += Icg + mass[k] * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        outputs["I_base"] = util.unassembleI(I_base)

        # Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
        # All dimensions for correlations based on mm, not meters.
        R_ave = 0.5 * (Rb + Rt)
        taper = np.minimum(Rb / Rt, Rt / Rb)
        nsec = twall.size
        mshell = rho * V_shell
        mshell_tot = np.sum(rho * V_shell)
        k_m = inputs["material_cost_rate"]  # 1.1 # USD / kg carbon steel plate
        k_f = inputs["labor_cost_rate"]  # 1.0 # USD / min labor
        k_p = inputs["painting_cost_rate"]  # USD / m^2 painting
        k_e = 0.064  # Industrial electricity rate $/kWh https://www.eia.gov/electricity/monthly/epm_table_grapher.php?t=epmt_5_6_a
        e_f = 15.9  # Electricity usage kWh/kg for steel
        e_fo = 26.9  # Electricity usage kWh/kg for stainless steel

        # Cost Step 1) Cutting flat plates for taper using plasma cutter
        cutLengths = 2.0 * np.sqrt((Rt - Rb) ** 2.0 + H ** 2.0)  # Factor of 2 for both sides
        # Cost Step 2) Rolling plates
        # Cost Step 3) Welding rolled plates into shells (set difficulty factor based on tapering with logistic function)
        theta_F = 4.0 - 3.0 / (1 + np.exp(-5.0 * (taper - 0.75)))
        # Cost Step 4) Circumferential welds to join cans together
        theta_A = 2.0

        # Labor-based expenses
        K_f = k_f * (
            manufacture.steel_cutting_plasma_time(cutLengths, twall)
            + manufacture.steel_rolling_time(theta_F, R_ave, twall)
            + manufacture.steel_butt_welding_time(theta_A, nsec, mshell_tot, cutLengths, twall)
            + manufacture.steel_butt_welding_time(theta_A, nsec, mshell_tot, 2 * np.pi * Rb[1:], twall[1:])
        )

        # Cost step 5) Painting- outside and inside
        theta_p = 2
        K_p = k_p * theta_p * 2 * (2 * np.pi * R_ave * H).sum()

        # Cost step 6) Outfitting with electricity usage
        K_o = np.sum(1.5 * k_m * (coeff - 1.0) * mshell)

        # Material cost with waste fraction, but without outfitting,
        K_m = 1.21 * np.sum(k_m * mshell)

        # Electricity usage
        K_e = np.sum(k_e * (e_f * mshell + e_fo * (coeff - 1.0) * mshell))

        # Assemble all costs for now
        tempSum = K_m + K_e + K_o + K_p + K_f

        # Capital cost share from BLS MFP by NAICS
        K_c = 0.118 * tempSum / (1.0 - 0.118)

        outputs["cost"] = tempSum + K_c


class TowerMass(om.ExplicitComponent):
    """
    Compute the tower and monopile masses, raw cost, and CG properties.

    Parameters
    ----------
    cylinder_mass : numpy array[nFull-1], [kg]
        Total cylinder mass
    cylinder_cost : float, [USD]
        Total cylinder cost
    cylinder_center_of_mass : float, [m]
        z position of center of mass of cylinder
    cylinder_section_center_of_mass : numpy array[nFull-1], [m]
        z position of center of mass of each can in the cylinder
    cylinder_I_base : numpy array[6], [kg*m**2]
        Mass moment of inertia of cylinder about base [xx yy zz xy xz yz]
    transition_piece_height : float, [m]
        Point mass height of transition piece above water line
    transition_piece_mass : float, [kg]
        Point mass of transition piece
    transition_piece_cost : float, [USD]
        Cost of transition piece
    gravity_foundation_mass : float, [kg]
        Extra mass of gravity foundation
    z_full : numpy array[nFull], [m]
        Parameterized locations along tower, linear lofting between
    d_full : numpy array[nFull], [m]
        diameter along tower

    Returns
    -------
    structural_cost : float, [USD]
        Total structural cost (tower+monopile)
    structural_mass : float, [kg]
        Total structural mass (tower+monopile)
    tower_raw_cost : float, [USD]
        Tower cost only
    tower_mass : float, [kg]
        Tower mass only
    tower_center_of_mass : float, [m]
        z-position of center of mass of tower
    tower_section_center_of_mass : numpy array[nFull-1], [m]
        z position of center of mass of each can in the tower
    tower_I_base : numpy array[6], [kg*m**2]
        Mass moment of inertia of tower about base [xx yy zz xy xz yz]
    monopile_mass : float, [kg]
        Mass of monopile from bottom of suction pile through transition piece
    monopile_cost : float, [USD]
        Total monopile cost

    """

    def initialize(self):
        self.options.declare("n_height")
        self.options.declare("n_refine")

    def setup(self):
        n_height = self.options["n_height"]
        n_refine = self.options["n_refine"]
        nFull = get_nfull(n_height, nref=n_refine)

        self.add_input("cylinder_mass", val=np.zeros(nFull - 1), units="kg")
        self.add_input("cylinder_cost", val=0.0, units="USD")
        self.add_input("cylinder_center_of_mass", val=0.0, units="m")
        self.add_input("cylinder_section_center_of_mass", val=np.zeros(nFull - 1), units="m")
        self.add_input("cylinder_I_base", np.zeros(6), units="kg*m**2")
        self.add_input("transition_piece_height", 0.0, units="m")
        self.add_input("transition_piece_mass", 0.0, units="kg")
        self.add_input("transition_piece_cost", 0.0, units="USD")
        self.add_input("gravity_foundation_mass", 0.0, units="kg")
        self.add_input("z_full", val=np.zeros(nFull), units="m")
        self.add_input("d_full", val=np.zeros(nFull), units="m")

        self.add_output("structural_cost", val=0.0, units="USD")
        self.add_output("structural_mass", val=0.0, units="kg")
        self.add_output("tower_cost", val=0.0, units="USD")
        self.add_output("tower_mass", val=0.0, units="kg")
        self.add_output("tower_center_of_mass", val=0.0, units="m")
        self.add_output("tower_section_center_of_mass", val=np.zeros(nFull - 1), units="m")
        self.add_output("tower_I_base", np.zeros(6), units="kg*m**2")
        self.add_output("monopile_mass", val=0.0, units="kg")
        self.add_output("monopile_cost", val=0.0, units="USD")
        self.add_output("transition_piece_I", np.zeros(6), units="kg*m**2")
        self.add_output("gravity_foundation_I", np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs):
        # Unpack inputs
        z = inputs["z_full"]
        d = inputs["d_full"]
        z_trans = inputs["transition_piece_height"]
        m_trans = inputs["transition_piece_mass"]
        m_grav = inputs["gravity_foundation_mass"]
        m_cyl = inputs["cylinder_mass"]

        outputs["structural_cost"] = inputs["cylinder_cost"] + inputs["transition_piece_cost"]
        outputs["structural_mass"] = m_cyl.sum() + m_trans + m_grav
        outputs["tower_center_of_mass"] = (
            inputs["cylinder_center_of_mass"] * m_cyl.sum() + m_trans * z_trans + m_grav * z[0]
        ) / (m_cyl.sum() + m_trans + m_grav)
        outputs["tower_section_center_of_mass"] = inputs["cylinder_section_center_of_mass"]

        outputs["monopile_mass"], dydx, dydxp, dydyp = util.interp_with_deriv(z_trans, z, np.r_[0.0, np.cumsum(m_cyl)])
        outputs["monopile_cost"] = (
            inputs["cylinder_cost"] * outputs["monopile_mass"] / m_cyl.sum() + inputs["transition_piece_cost"]
        )
        outputs["monopile_mass"] += m_trans + m_grav

        outputs["tower_cost"] = outputs["structural_cost"] - outputs["monopile_cost"]
        outputs["tower_mass"] = outputs["structural_mass"] - outputs["monopile_mass"]
        outputs["tower_I_base"] = inputs["cylinder_I_base"]
        outputs["tower_I_base"][:2] += m_trans * (z_trans - z[0]) ** 2

        # Mass properties for transition piece and gravity foundation
        itrans = util.find_nearest(z, z_trans)
        r_trans = 0.5 * d[itrans]
        r_grav = 0.5 * d[0]
        I_trans = m_trans * r_trans ** 2.0 * np.r_[0.5, 0.5, 1.0, np.zeros(3)]  # shell
        I_grav = m_grav * r_grav ** 2.0 * np.r_[0.25, 0.25, 0.5, np.zeros(3)]  # disk
        outputs["transition_piece_I"] = I_trans
        outputs["gravity_foundation_I"] = I_grav


class TurbineMass(om.ExplicitComponent):
    """
    Compute the turbine mass, center of mass, and mass moment of inertia.

    Parameters
    ----------
    hub_height : float, [m]
        Hub-height
    rna_mass : float, [kg]
        Total tower mass
    rna_I : numpy array[6], [kg*m**2]
        Mass moment of inertia of RNA about tower top [xx yy zz xy xz yz]
    rna_cg : numpy array[3], [m]
        xyz-location of RNA cg relative to tower top
    tower_mass : float, [kg]
        Total tower mass (not including monopile)
    monopile_mass : float, [kg]
        Monopile mass
    tower_center_of_mass : float, [m]
        z-position of center of mass of tower
    tower_I_base : numpy array[6], [kg*m**2]
        Mass moment of inertia of tower about base [xx yy zz xy xz yz]

    Returns
    -------
    turbine_mass : float, [kg]
        Total mass of tower+rna
    turbine_center_of_mass : numpy array[3], [m]
        xyz-position of tower+rna center of mass
    turbine_I_base : numpy array[6], [kg*m**2]
        mass moment of inertia of tower about base [xx yy zz xy xz yz]

    """

    def setup(self):
        self.add_input("hub_height", val=0.0, units="m")
        self.add_input("rna_mass", val=0.0, units="kg")
        self.add_input("rna_I", np.zeros(6), units="kg*m**2")
        self.add_input("rna_cg", np.zeros(3), units="m")
        self.add_input("tower_mass", val=0.0, units="kg")
        self.add_input("monopile_mass", val=0.0, units="kg")
        self.add_input("tower_center_of_mass", val=0.0, units="m")
        self.add_input("tower_I_base", np.zeros(6), units="kg*m**2")

        self.add_output("turbine_mass", val=0.0, units="kg")
        self.add_output("turbine_center_of_mass", val=np.zeros(3), units="m")
        self.add_output("turbine_I_base", np.zeros(6), units="kg*m**2")

        self.declare_partials(
            "turbine_I_base", ["hub_height", "rna_I", "rna_cg", "rna_mass", "tower_I_base"], method="fd"
        )
        self.declare_partials(
            "turbine_center_of_mass",
            ["hub_height", "monopile_mass", "rna_cg", "rna_mass", "tower_center_of_mass", "tower_mass"],
            method="fd",
        )
        self.declare_partials("turbine_mass", ["monopile_mass", "rna_mass", "tower_mass"], val=1.0)

    def compute(self, inputs, outputs):
        outputs["turbine_mass"] = inputs["rna_mass"] + inputs["tower_mass"]

        cg_rna = inputs["rna_cg"] + np.r_[0.0, 0.0, inputs["hub_height"]]
        cg_tower = np.r_[0.0, 0.0, inputs["tower_center_of_mass"]]
        outputs["turbine_center_of_mass"] = (inputs["rna_mass"] * cg_rna + inputs["tower_mass"] * cg_tower) / outputs[
            "turbine_mass"
        ]

        R = cg_rna
        I_tower = util.assembleI(inputs["tower_I_base"])
        I_rna = util.assembleI(inputs["rna_I"]) + inputs["rna_mass"] * (np.dot(R, R) * np.eye(3) - np.outer(R, R))
        outputs["turbine_I_base"] = util.unassembleI(I_tower + I_rna)



"""This package contains simulation logic shared across several modules."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from .vessel_logic import (  # shuttle_items_to_queue
    stabilize,
    position_onsite,
    jackdown_if_required,
    shuttle_items_to_queue,
    prep_for_site_operations,
    get_list_of_items_from_port,
)



#!/usr/bin/env python
# encoding: utf-8
"""
environment.py

Created by Andrew Ning on 2012-01-20.
Copyright (c) NREL. All rights reserved.
"""

from __future__ import print_function

import numpy as np
import openmdao.api as om
from scipy.optimize import brentq
from wisdem.commonse.constants import gravity

# TODO CHECK

# -----------------
#  Base Components
# -----------------


class WindBase(om.ExplicitComponent):
    """
    Base component for wind speed/direction

    Parameters
    ----------
    Uref : float, [m/s]
        reference wind speed (usually at hub height)
    zref : float, [m]
        corresponding reference height
    z : numpy array[npts], [m]
        'heights where wind speed should be computed
    z0 : float, [m]
        bottom of wind profile (height of ground/sea)

    Returns
    -------
    U : numpy array[npts], [m/s]
        magnitude of wind speed at each z location

    """

    def initialize(self):
        self.options.declare("nPoints")

    def setup(self):
        npts = self.options["nPoints"]

        self.add_input("Uref", 0.0, units="m/s")
        self.add_input("zref", 0.0, units="m")
        self.add_input("z", np.zeros(npts), units="m")
        self.add_input("z0", 0.0, units="m")

        self.add_output("U", np.zeros(npts), units="m/s")


class WaveBase(om.ExplicitComponent):
    """
    Base component for wave speed/direction

    Parameters
    ----------
    rho_water : float, [kg/m**3]
        water density
    z : numpy array[npts], [m]
        heights where wave speed should be computed
    z_surface : float, [m]
        vertical location of water surface
    z_floor : float, [m]
        vertical location of sea floor

    Returns
    -------
    U : numpy array[npts], [m/s]
        horizontal wave velocity at each z location
    W : numpy array[npts], [m/s]
        vertical wave velocity at each z location
    V : numpy array[npts], [m/s]
        total wave velocity at each z location
    A : numpy array[npts], [m/s**2]
        horizontal wave acceleration at each z location
    p : numpy array[npts], [N/m**2]
        pressure oscillation at each z location

    """

    def initialize(self):
        self.options.declare("nPoints")

    def setup(self):
        npts = self.options["nPoints"]

        self.add_input("rho_water", 0.0, units="kg/m**3")
        self.add_input("z", np.zeros(npts), units="m")
        self.add_input("z_surface", 0.0, units="m")
        self.add_input("z_floor", 0.0, units="m")

        self.add_output("U", np.zeros(npts), units="m/s")
        self.add_output("W", np.zeros(npts), units="m/s")
        self.add_output("V", np.zeros(npts), units="m/s")
        self.add_output("A", np.zeros(npts), units="m/s**2")
        self.add_output("p", np.zeros(npts), units="N/m**2")

    def compute(self, inputs, outputs):
        """default to no waves"""
        n = len(inputs["z"])
        outputs["U"] = np.zeros(n)
        outputs["W"] = np.zeros(n)
        outputs["V"] = np.zeros(n)
        outputs["A"] = np.zeros(n)
        outputs["p"] = np.zeros(n)
        # outputs['U0'] = 0.
        # outputs['A0'] = 0.


# -----------------------
#  Subclassed Components
# -----------------------


class PowerWind(WindBase):
    """
    Power-law profile wind.
    Any nodes must not cross z0, and if a node is at z0 it must stay at that point.
    Otherwise gradients crossing the boundary will be wrong.

    Parameters
    ----------
    Uref : float, [m/s]
        reference wind speed (usually at hub height)
    zref : float, [m]
        corresponding reference height
    z : numpy array[npts], [m]
        'heights where wind speed should be computed
    z0 : float, [m]
        bottom of wind profile (height of ground/sea)
    shearExp : float
        shear exponent

    Returns
    -------
    U : numpy array[npts], [m/s]
        magnitude of wind speed at each z location

    """

    def setup(self):
        super(PowerWind, self).setup()

        # parameters
        self.add_input("shearExp", 0.0)

        self.declare_partials("U", ["Uref", "z", "zref"])

    def compute(self, inputs, outputs):

        # rename
        z = inputs["z"]
        if isinstance(z, float) or isinstance(z, np.float_):
            z = np.array([z])
        zref = float(inputs["zref"])
        z0 = float(inputs["z0"])

        # velocity
        idx = z > z0
        outputs["U"] = np.zeros(self.options["nPoints"])
        outputs["U"][idx] = inputs["Uref"] * ((z[idx] - z0) / (zref - z0)) ** inputs["shearExp"]

        # # add small cubic spline to allow continuity in gradient
        # k = 0.01  # fraction of profile with cubic spline
        # zsmall = z0 + k*(zref - z0)

        # self.spline = CubicSpline(x1=z0, x2=zsmall, f1=0.0, f2=Uref*k**shearExp,
        #     g1=0.0, g2=Uref*k**shearExp*shearExp/(zsmall - z0))

        # idx = np.logical_and(z > z0, z < zsmall)
        # self.U[idx] = self.spline.eval(z[idx])

        # self.zsmall = zsmall
        # self.k = k

    def compute_partials(self, inputs, J):

        # rename
        z = inputs["z"]
        if isinstance(z, float) or isinstance(z, np.float_):
            z = np.array([z])
        zref = inputs["zref"]
        z0 = inputs["z0"]
        shearExp = inputs["shearExp"]
        idx = z > z0
        npts = self.options["nPoints"]

        U = np.zeros(npts)
        U[idx] = inputs["Uref"] * ((z[idx] - z0) / (zref - z0)) ** inputs["shearExp"]

        # gradients
        dU_dUref = np.zeros(npts)
        dU_dz = np.zeros(npts)
        dU_dzref = np.zeros(npts)

        dU_dUref[idx] = U[idx] / inputs["Uref"]
        dU_dz[idx] = U[idx] * shearExp / (z[idx] - z0)
        dU_dzref[idx] = -U[idx] * shearExp / (zref - z0)

        J["U", "Uref"] = dU_dUref
        J["U", "z"] = np.diag(dU_dz)
        J["U", "zref"] = dU_dzref
        # TODO still missing several partials? This is what was in the original code though...

        # # cubic spline region
        # idx = np.logical_and(z > z0, z < zsmall)

        # # d w.r.t z
        # dU_dz[idx] = self.spline.eval_deriv(z[idx])

        # # d w.r.t. Uref
        # df2_dUref = k**shearExp
        # dg2_dUref = k**shearExp*shearExp/(zsmall - z0)
        # dU_dUref[idx] = self.spline.eval_deriv_inputs(z[idx], 0.0, 0.0, 0.0, df2_dUref, 0.0, dg2_dUref)

        # # d w.r.t. zref
        # dx2_dzref = k
        # dg2_dzref = -Uref*k**shearExp*shearExp/k/(zref - z0)**2
        # dU_dzref[idx] = self.spline.eval_deriv_params(z[idx], 0.0, dx2_dzref, 0.0, 0.0, 0.0, dg2_dzref)


class LogWind(WindBase):
    """
    Logarithmic-profile wind

    Parameters
    ----------
    Uref : float, [m/s]
        reference wind speed (usually at hub height)
    zref : float, [m]
        corresponding reference height
    z : numpy array[npts], [m]
        'heights where wind speed should be computed
    z0 : float, [m]
        bottom of wind profile (height of ground/sea)
    z_roughness : float, [mm]
        surface roughness length

    Returns
    -------
    U : numpy array[npts], [m/s]
        magnitude of wind speed at each z location

    """

    def setup(self):
        super(LogWind, self).setup()

        # parameters
        self.add_input("z_roughness", 1e-3, units="mm")

        self.declare_partials("U", ["Uref", "z", "zref"])

    def compute(self, inputs, outputs):

        # rename
        z = inputs["z"]
        if isinstance(z, float) or isinstance(z, np.float_):
            z = np.array([z])
        zref = float(inputs["zref"])
        z0 = float(inputs["z0"])
        z_roughness = float(inputs["z_roughness"]) / 1e3  # convert to m

        # find velocity
        idx = np.where(z - z0 > z_roughness)[0]
        outputs["U"] = np.zeros_like(z)
        outputs["U"][idx] = inputs["Uref"] * np.log((z[idx] - z0) / z_roughness) / np.log((zref - z0) / z_roughness)

    def compute_partials(self, inputs, J):

        # rename
        z = inputs["z"]
        if isinstance(z, float) or isinstance(z, np.float_):
            z = np.array([z])
        zref = float(inputs["zref"])
        z0 = float(inputs["z0"])
        z_roughness = float(inputs["z_roughness"]) / 1e3
        Uref = float(inputs["Uref"])
        npts = self.options["nPoints"]

        dU_dUref = np.zeros(npts)
        dU_dz_diag = np.zeros(npts)
        dU_dzref = np.zeros(npts)

        idx = np.where(z - z0 > z_roughness)[0]
        lt = np.log((z[idx] - z0) / z_roughness)
        lb = np.log((zref - z0) / z_roughness)
        dU_dUref[idx] = lt / lb
        dU_dz_diag[idx] = Uref / lb / (z[idx] - z0)
        dU_dzref[idx] = -Uref * lt / np.log((zref - z0) / z_roughness) ** 2 / (zref - z0)

        J["U", "Uref"] = dU_dUref
        J["U", "z"] = np.diag(dU_dz_diag)
        J["U", "zref"] = dU_dzref


class LinearWaves(WaveBase):
    """
    Linear (Airy) wave theory

    Parameters
    ----------
    rho_water : float, [kg/m**3]
        water density
    z : numpy array[npts], [m]
        heights where wave speed should be computed
    z_surface : float, [m]
        vertical location of water surface
    z_floor : float, [m]
        vertical location of sea floor
    Hsig_wave : float, [m]
        Maximum wave height (crest-to-trough)
    Tsig_wave : float, [s]
        period of maximum wave height

    Returns
    -------
    U : numpy array[npts], [m/s]
        horizontal wave velocity at each z location
    W : numpy array[npts], [m/s]
        vertical wave velocity at each z location
    V : numpy array[npts], [m/s]
        total wave velocity at each z location
    A : numpy array[npts], [m/s**2]
        horizontal wave acceleration at each z location
    p : numpy array[npts], [N/m**2]
        pressure oscillation at each z location
    phase_speed : float, [m/s]
        Phase speed of wave

    """

    def setup(self):
        super(LinearWaves, self).setup()

        # variables
        self.add_input("Uc", 0.0, units="m/s", desc="mean current speed")

        # parameters
        self.add_input("Hsig_wave", 0.0, units="m")
        self.add_input("Tsig_wave", 0.0, units="s")

        # For Ansys AQWA connection
        self.add_output("phase_speed", val=0.0, units="m/s")

        self.declare_partials("U", ["Uc", "z"])
        self.declare_partials("V", ["Uc", "z"])
        self.declare_partials("W", ["Uc", "z"])
        self.declare_partials("A", ["Uc", "z"])
        self.declare_partials("p", ["Uc", "z"])

    def compute(self, inputs, outputs):
        super(LinearWaves, self).compute(inputs, outputs)

        # water depth
        z_floor = inputs["z_floor"]
        if z_floor > 0.0:
            z_floor *= -1.0
        d = inputs["z_surface"] - z_floor
        # Use zero entries if there is no depth and no water
        if d == 0.0:
            return

        # design wave height
        h = inputs["Hsig_wave"]

        # circular frequency
        omega = 2.0 * np.pi / inputs["Tsig_wave"]

        # compute wave number from dispersion relationship
        k = brentq(lambda k: omega ** 2 - gravity * k * np.tanh(d * k), 0, 1e3 * omega ** 2 / gravity)
        self.k = k
        outputs["phase_speed"] = omega / k

        # zero at surface
        z_rel = inputs["z"] - inputs["z_surface"]

        # Amplitude
        a = 0.5 * h

        # maximum velocity
        outputs["U"] = a * omega * np.cosh(k * (z_rel + d)) / np.sinh(k * d) + inputs["Uc"]
        outputs["W"] = -a * omega * np.sinh(k * (z_rel + d)) / np.sinh(k * d)
        outputs["V"] = np.sqrt(outputs["U"] ** 2.0 + outputs["W"] ** 2.0)
        # outputs['U0'] = a*omega*np.cosh(k*(0. + d))/np.sinh(k*d) + inputs['Uc']

        # acceleration
        outputs["A"] = (outputs["U"] - inputs["Uc"]) * omega
        # outputs['A0'] = (outputs['U0'] - inputs['Uc']) * omega

        # Pressure oscillation is just sum of static and dynamic contributions
        # Hydrostatic is simple rho * g * z
        # Dynamic is from standard solution to Airy (Potential Flow) Wave theory
        # Full pressure would also include standard dynamic head (0.5*rho*V^2)
        outputs["p"] = inputs["rho_water"] * gravity * (a * np.cosh(k * (z_rel + d)) / np.cosh(k * d) - z_rel)

        # check heights
        idx = np.logical_or(inputs["z"] < z_floor, inputs["z"] > inputs["z_surface"])
        outputs["U"][idx] = 0.0
        outputs["W"][idx] = 0.0
        outputs["V"][idx] = 0.0
        outputs["A"][idx] = 0.0
        outputs["p"][idx] = 0.0

    def compute_partials(self, inputs, J):
        outputs = {}
        self.compute(inputs, outputs)

        # rename
        z_floor = inputs["z_floor"]
        if z_floor > 0.0:
            z_floor *= -1.0
        z = inputs["z"]
        d = inputs["z_surface"] - z_floor
        h = inputs["Hsig_wave"]
        omega = 2.0 * np.pi / inputs["Tsig_wave"]
        k = self.k
        z_rel = z - inputs["z_surface"]

        # Amplitude
        a = 0.5 * h

        # derivatives
        dU_dz = h / 2.0 * omega * np.sinh(k * (z_rel + d)) / np.sinh(k * d) * k
        dU_dUc = np.ones_like(z)
        dW_dz = -h / 2.0 * omega * np.cosh(k * (z_rel + d)) / np.sinh(k * d) * k
        dV_dz = 0.5 / outputs["V"] * (2 * outputs["U"] * dU_dz + 2 * outputs["W"] * dW_dz)
        dV_dUc = 0.5 / outputs["V"] * (2 * outputs["U"] * dU_dUc)
        dA_dz = omega * dU_dz
        dA_dUc = 0.0  # omega*dU_dUc
        dp_dz = inputs["rho_water"] * gravity * (a * np.sinh(k * (z_rel + d)) * k / np.cosh(k * d) - 1.0)

        idx = np.logical_or(z < z_floor, z > inputs["z_surface"])
        dU_dz[idx] = 0.0
        dW_dz[idx] = 0.0
        dV_dz[idx] = 0.0
        dA_dz[idx] = 0.0
        dp_dz[idx] = 0.0
        dU_dUc[idx] = 0.0
        dV_dUc[idx] = 0.0

        # dU0 = np.zeros((1,npts))
        # dA0 = omega * dU0

        J["U", "z"] = np.diag(dU_dz)
        J["U", "Uc"] = dU_dUc
        J["W", "z"] = np.diag(dW_dz)
        J["W", "Uc"] = 0.0
        J["V", "z"] = np.diag(dV_dz)
        J["V", "Uc"] = 0.0
        J["A", "z"] = np.diag(dA_dz)
        J["A", "Uc"] = 0.0
        J["p", "z"] = np.diag(dp_dz)
        J["p", "Uc"] = 0.0
        # J['U0', 'z'] = dU0
        # J['U0', 'Uc'] = 1.0
        # J['A0', 'z'] = dA0
        # J['A0', 'Uc'] = 1.0


class TowerSoil(om.ExplicitComponent):
    """
    Soil stiffness method from Arya, Suresh C., Michael W. O'Neill, and George Pincus.
    Design of structures and foundations for vibrating machines. Gulf Pub Co, 1979.

    Parameters
    ----------
    d0 : float, [m]
        diameter of base of tower
    depth : float, [m]
        depth of foundation in the soil
    G : float, [Pa]
        shear modulus of soil
    nu : float
        Poisson's ratio of soil
    k_usr : numpy array[6], [N/m]
        User overrides of stiffness values. Use positive values and for rigid
        use np.inf. Order is x, theta_x, y, theta_y, z, theta_z

    Returns
    -------
    k : numpy array[6], [N/m]
        Spring stiffness (x, theta_x, y, theta_y, z, theta_z)
    """

    def initialize(self):
        self.options.declare("npts", default=1)

    def setup(self):
        npts = self.options["npts"]

        # variable
        self.add_input("d0", 1.0, units="m")
        self.add_input("depth", 0.0, units="m")

        # inputeter
        self.add_input("G", 140e6, units="Pa")
        self.add_input("nu", 0.4)
        self.add_input("k_usr", -1 * np.ones(6), units="N/m")

        self.add_output("z_k", np.zeros(npts), units="N/m")
        self.add_output("k", np.zeros((npts, 6)), units="N/m")

        self.declare_partials("k", ["d0", "depth"])

    def compute(self, inputs, outputs):

        G = inputs["G"]
        nu = inputs["nu"]
        h = np.linspace(inputs["depth"], 0.0, self.options["npts"])
        r0 = 0.5 * inputs["d0"]

        # vertical
        eta = 1.0 + 0.6 * (1.0 - nu) * h / r0
        k_z = 4 * G * r0 * eta / (1.0 - nu)

        # horizontal
        eta = 1.0 + 0.55 * (2.0 - nu) * h / r0
        k_x = 32.0 * (1.0 - nu) * G * r0 * eta / (7.0 - 8.0 * nu)

        # rocking
        eta = 1.0 + 1.2 * (1.0 - nu) * h / r0 + 0.2 * (2.0 - nu) * (h / r0) ** 3
        k_thetax = 8.0 * G * r0 ** 3 * eta / (3.0 * (1.0 - nu))

        # torsional
        k_phi = 16.0 * G * r0 ** 3 * np.ones(h.size) / 3.0

        outputs["k"] = np.c_[k_x, k_thetax, k_x, k_thetax, k_z, k_phi]
        outputs["z_k"] = -h
        ind = np.nonzero(inputs["k_usr"] >= 0.0)[0]
        outputs["k"][:, ind] = inputs["k_usr"][np.newaxis, ind]

    def compute_partials(self, inputs, J):

        G = inputs["G"]
        nu = inputs["nu"]
        h = np.linspace(inputs["depth"], 0.0, self.options["npts"])
        r0 = 0.5 * inputs["d0"]

        # vertical
        eta = 1.0 + 0.6 * (1.0 - nu) * h / r0
        deta_dr0 = -0.6 * (1.0 - nu) * h / r0 ** 2
        dkz_dr0 = 4 * G / (1.0 - nu) * (eta + r0 * deta_dr0)

        deta_dh = 0.6 * (1.0 - nu) / r0
        dkz_dh = 4 * G * r0 / (1.0 - nu) * deta_dh

        # horizontal
        eta = 1.0 + 0.55 * (2.0 - nu) * h / r0
        deta_dr0 = -0.55 * (2.0 - nu) * h / r0 ** 2
        dkx_dr0 = 32.0 * (1.0 - nu) * G / (7.0 - 8.0 * nu) * (eta + r0 * deta_dr0)

        deta_dh = 0.55 * (2.0 - nu) / r0
        dkx_dh = 32.0 * (1.0 - nu) * G * r0 / (7.0 - 8.0 * nu) * deta_dh

        # rocking
        eta = 1.0 + 1.2 * (1.0 - nu) * h / r0 + 0.2 * (2.0 - nu) * (h / r0) ** 3
        deta_dr0 = -1.2 * (1.0 - nu) * h / r0 ** 2 - 3 * 0.2 * (2.0 - nu) * (h / r0) ** 3 / r0
        dkthetax_dr0 = 8.0 * G / (3.0 * (1.0 - nu)) * (3 * r0 ** 2 * eta + r0 ** 3 * deta_dr0)

        deta_dh = 1.2 * (1.0 - nu) / r0 + 3 * 0.2 * (2.0 - nu) * (1.0 / r0) ** 3 * h ** 2
        dkthetax_dh = 8.0 * G * r0 ** 3 / (3.0 * (1.0 - nu)) * deta_dh

        # torsional
        dkphi_dr0 = 16.0 * G * 3 * r0 ** 2 / 3.0
        dkphi_dh = 0.0

        dk_dr0 = np.c_[dkx_dr0, dkthetax_dr0, dkx_dr0, dkthetax_dr0, dkz_dr0, dkphi_dr0]
        # dk_dr0[inputs['rigid']] = 0.0
        dk_dh = np.c_[dkx_dh, dkthetax_dh, dkx_dh, dkthetax_dh, dkz_dh, dkphi_dh]
        # dk_dh[inputs['rigid']] = 0.0

        J["k", "d0"] = 0.5 * dk_dr0
        J["k", "depth"] = dk_dh
        ind = np.nonzero(inputs["k_usr"] >= 0.0)[0]
        J["k", "d0"][:, ind] = 0.0
        J["k", "depth"][:, ind] = 0.0



import openmdao.api as om
import numpy as np
from scipy.optimize import minimize

# Application factor to include ring/housing/carrier weight
Kr = 0.4

# -----------------------------------


def V_planetary(U, B, K):
    sunU = 0.5 * U - 1.0
    V = (
        1.0 / U
        + 1.0 / U / B
        + 1.0 / B / sunU
        + sunU
        + sunU ** 2
        + K * (U - 1.0) ** 2 / B
        + K * (U - 1.0) ** 2 / B / sunU
    )
    return V


# -----------------------------------


def V_parallel(U):
    V = 1.0 + 1.0 / U + U + U ** 2
    return V


# -----------------------------------
def volumeEEP(x, n_planets, torque, Kr1=Kr, Kr2=Kr):
    # Safety factor?
    Kgamma = [1.1 if m < 5 else 1.35 for m in n_planets]

    # Individual stage torques
    Q_stage = torque / np.cumprod(x)

    # Volume
    V = (
        Q_stage[0] * Kgamma[0] * V_planetary(x[0], n_planets[0], Kr1)
        + Q_stage[1] * Kgamma[1] * V_planetary(x[1], n_planets[1], Kr2)
        + Q_stage[2] * V_parallel(x[2]) / np.prod(x)
    )
    return 2 * V


# -----------------------------------


def volumeEPP(x, n_planets, torque, Kr1=Kr):
    # Safety factor?
    Kgamma = [1.1 if m < 5 else 1.35 for m in n_planets]

    # Individual stage torques
    Q_stage = torque / np.cumprod(x)

    V = (
        Q_stage[0] * Kgamma[0] * V_planetary(x[0], n_planets[0], Kr1)
        + Q_stage[1] * V_parallel(x[1]) / np.prod(x[:2])
        + Q_stage[2] * V_parallel(x[2]) / np.prod(x)
    )
    return 2 * V


# -----------------------------------


class Gearbox(om.ExplicitComponent):
    """
    The gearbox design follows the general approach of the previous DriveSE implementation, however
    with code improvements, the results will likely be different than prior versions.  The gearbox is
    assumed to have 3 stages, with the user specifying a configuration code of either "EEP" or "EPP",
    with the "E" representing epicyclic (planetary) gear stages and "P" representing parallel gear stages.
    For the epicyclic stages, the user also has to specify the number of planets, so the EEP input would
    require something like [3, 3, 0] and EPP would require [3, 0, 0].  The user also specifies the overall
    target gear ratio, and then DrivetrainSE conducts a mass minimization of the three stage ratios that
    meet the target and minimize the overall mass.

    Parameters
    ----------
    gear_configuration : string
        3-letter string of Es or Ps to denote epicyclic or parallel gear configuration
    n_planets : numpy array[3]
        number of planets in each stage
    gear_ratio : float
        overall gearbox speedup ratio
    D_rotor : float, [m]
        rotor diameter
    Q_rotor : float, [N*m]
        rotor torque at rated power
    s_gearbox : float, [m]
        gearbox position along x-axis

    Returns
    -------
    stage_masses : numpy array[3], [kg]
        individual gearbox stage gearbox_masses
    gearbox_mass : float, [kg]
        overall component mass
    gearbox_cm : numpy array[3], [m]
        Gearbox center of mass [x,y,z] measure along shaft from bedplate
    gearbox_I : numpy array[3], [kg*m**2]
        Gearbox mass moments of inertia [Ixx, Iyy, Izz] around its center of mass
    L_gearbox : float, [m]
        length of gearbox
    H_gearbox : float, [m]
        height of gearbox
    D_gearbox : float, [m]
        diameter of gearbox

    """

    def initialize(self):
        self.options.declare("direct_drive", default=True)

    def setup(self):
        self.add_discrete_input("gear_configuration", val="eep")
        # self.add_discrete_input('shaft_factor', val='normal')
        self.add_discrete_input("planet_numbers", val=[3, 3, 0])
        self.add_input("gear_ratio", val=1.0)
        self.add_input("rotor_diameter", val=0.0, units="m")
        self.add_input("rated_torque", val=0.0, units="N*m")
        self.add_input("machine_rating", val=0.0, units="kW")

        self.add_output("stage_ratios", val=np.zeros(3))
        self.add_output("gearbox_mass", 0.0, units="kg")
        self.add_output("gearbox_I", np.zeros(3), units="kg*m**2")
        self.add_output("L_gearbox", 0.0, units="m")
        self.add_output("D_gearbox", 0.0, units="m")
        self.add_output("carrier_mass", 0.0, units="kg")
        self.add_output("carrier_I", np.zeros(3), units="kg*m**2")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        if self.options["direct_drive"]:
            outputs["stage_ratios"] = np.zeros(3)
            outputs["gearbox_mass"] = outputs["D_gearbox"] = outputs["L_gearbox"] = 0.0
            outputs["gearbox_I"] = np.zeros(3)
            return

        # Unpack inputs
        config = discrete_inputs["gear_configuration"]
        # shaft_factor = discrete_inputs['shaft_factor']
        n_planets = np.maximum(1.0, np.array(discrete_inputs["planet_numbers"]))
        gear_ratio = inputs["gear_ratio"]
        D_rotor = inputs["rotor_diameter"]
        torque = inputs["rated_torque"]
        rating = inputs["machine_rating"]

        # Known configuration checks
        if not config.lower() in ["eep", "eep_2", "eep_3", "epp"]:
            raise ValueError("Invalid value for gearbox_configuration.  Must be one of: eep, eep_2, eep_3, epp")
        n_stage = 3

        # Optimize stage ratios

        # Use double sided constraints to hack inequality constraints as COBYLA seems to work better than SLSQP here
        def constr1(x, ratio):
            return np.prod(x) - ratio

        def constr2(x, ratio):
            return ratio - np.prod(x)

        x0 = gear_ratio ** (1.0 / n_stage) * np.ones(n_stage)
        bounds = [[2.01, 20.0], [2.01, 20.0], [2.01, 20.0]]
        const = [{}, {}]
        const[0]["type"] = "ineq"
        const[0]["fun"] = constr1
        const[0]["args"] = [gear_ratio]
        const[1]["type"] = "ineq"
        const[1]["fun"] = constr2
        const[1]["args"] = [gear_ratio]
        method = "cobyla"
        tol = 1e-3

        if config.lower() == "eep":
            bounds[2][0] = 1.0
            result = minimize(
                lambda x: volumeEEP(x, n_planets, torque),
                x0,
                method=method,  # bounds=bounds,
                tol=tol,
                constraints=const,
                options={"maxiter": 100},
            )
            ratio_stage = result.x

        elif config == "eep_3":
            # fixes last stage ratio at 3
            const[0]["args"] = const[1]["args"] = [gear_ratio / 3.0]
            bounds[2][0] = 1.0
            result = minimize(
                lambda x: volumeEEP(np.r_[x, 3.0], n_planets, torque),
                x0[:2],
                method=method,  # bounds=bounds,
                tol=tol,
                constraints=const,
                options={"maxiter": 100},
            )
            ratio_stage = np.r_[result.x, 3.0]

        elif config == "eep_2":
            # fixes final stage ratio at 2
            const[0]["args"] = const[1]["args"] = [gear_ratio / 2.0]
            bounds[2][0] = 1.0
            result = minimize(
                lambda x: volumeEEP(np.r_[x, 2.0], n_planets, torque),
                x0[:2],
                method=method,  # bounds=bounds,
                tol=tol,
                constraints=const,
                options={"maxiter": 100},
            )
            ratio_stage = np.r_[result.x, 2.0]

        elif config == "epp":
            bounds[1][0] = 1.0
            bounds[2][0] = 1.0
            result = minimize(
                lambda x: volumeEPP(x, n_planets, torque),
                x0,
                method=method,  # bounds=bounds,
                tol=tol,
                constraints=const,
                options={"maxiter": 100},
            )
            ratio_stage = result.x

        # Get final volume
        if config.lower().find("eep") >= 0:
            vol = volumeEEP(ratio_stage, n_planets, torque)
        else:
            vol = volumeEPP(ratio_stage, n_planets, torque)

        # Cumulative Kfactor scaling based on values reported in Nejad's paper even though that was probably done with the buggy version of DriveSE
        K = np.mean([48.82 / 15216.504, 53.69 / 17401.453])

        # Shaft length factor
        Kshaft = 1.0  # if shaft_factor == 'normal' else 1.25

        # All factors into the mass
        m_gearbox = K * Kshaft * vol.sum()

        # Other gearbox elements that are just estimates: shrink disc and carrier
        m_shrink_disc = rating / 3.0
        m_carrier = 8e3
        m_gearbox += m_shrink_disc + m_carrier

        # calculate mass properties
        L_gearbox = 0.012 * D_rotor
        R_gearbox = 0.5 * 0.75 * 0.015 * D_rotor

        I = np.zeros(3)
        I[0] = 0.5 * R_gearbox ** 2
        I[1:] = (1.0 / 12.0) * (3 * R_gearbox ** 2 + L_gearbox ** 2)

        # Store outputs
        outputs["stage_ratios"] = ratio_stage
        outputs["gearbox_mass"] = m_gearbox
        outputs["gearbox_I"] = I * m_gearbox
        outputs["D_gearbox"] = 2 * R_gearbox
        outputs["L_gearbox"] = L_gearbox
        outputs["carrier_mass"] = m_shrink_disc + m_carrier
        outputs["carrier_I"] = outputs["carrier_mass"] * I[0] * np.array([1.0, 0.5, 0.5])  # Solid disk
